from __future__ import annotations

import csv
import glob
import json
import logging
import os
from collections import Counter
from collections import defaultdict
from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Union

import numpy as np
import pandas as pd
import xxhash

logging.basicConfig(format='%(asctime)s %(message)s', level=logging.INFO)

pro_tips = [
    'OutRank can construct subfeatures; features based on subspaces. Example command argument is: --subfeature_mapping "feature_a->feature_b;feature_c<->feature_d;feature_c<->feature_e"',
    'Heuristic MI-numba-randomized seems like the best of both worlds! (speed + performance).',
    'Heuristic surrogate-lr performs cross-validation (internally), keep that in mind!',
    'Consider running OutRank on a smaller data sample first, might be enough (--subsampling = a lot).',
    'There are two types of combinations supported; unsupervised pairwise ranking (redundancies- --target_ranking_only=False), and supervised combinations - (--interaction_order > 1)',
    'Visualization part also includes clustering - this might be very insightful!',
    'By default OutRank includes feature cardinality and coverage in feature names (card; cov)',
    'Intermediary checkpoints (tmp_checkpoint.tsv) might already give you insights during longer runs.',
    'In theory, you can rank redundancies of combined features (--interaction_order AND --target_ranking_only=False).',
    'Give it as many threads as physically possible (--num_threads).',
    'You can speed up ranking by diminishing feature buffer size (--combination_number_upper_bound determines how many ranking computations per batch will be considered). This, and --subsampling are very powerful together.',
    'Want to rank feature transformations, but not sure which ones to choose? --transformers=default should serve as a solid baseline (common DS transformations included).',
    'Your target can be any feature! (explaining one feature with others)',
    'OutRank uses HyperLogLog for cardinality estimation - this is also a potential usecase (understanding cardinalities across different data sets).',
    'Each feature is named as featureName(cardinality, coverage in percents) in the final files.',
    'You can generate candidate feature transformation ranges (fw) by using --task=feature_summary_transformers.',
]


def write_json_dump_to_file(args: Any, config_name: str) -> None:

    out_content = json.dumps(args.__dict__)
    with open(config_name, 'w') as out_config:
        out_config.write(out_content)


def internal_hash(input_obj: str) -> str:
    """A generic internal hash used throughout ranking procedure - let's hardcode seed here for sure"""
    return xxhash.xxh32(input_obj, seed=20141025).hexdigest()


@dataclass
class DatasetInformationStorage:
    """A generic class for holding properties of a given type of dataset"""

    data_path: str
    column_names: list[str]
    column_types: set[str]
    col_delimiter: str | None
    encoding: str
    fw_map: dict[str, str] | None


@dataclass
class NumericFeatureSummary:
    """A generic class storing numeric feature statistics"""

    feature_name: str
    minimum: float
    maximum: float
    median: float
    num_unique: int


@dataclass
class NominalFeatureSummary:
    """A generic class storing numeric feature statistics"""

    feature_name: str
    num_unique: int


@dataclass
class BatchRankingSummary:
    """A generic class representing batched ranking results"""

    triplet_scores: list[tuple[str, str, float]]
    step_times: dict[str, Any]


def display_random_tip() -> None:
    TIP_CONTENT = np.random.choice(pro_tips)
    tip_core = f"""
=====>
Random tip: {TIP_CONTENT}
=====>
    """

    print(tip_core)


def get_dataset_info(args: Any):
    if args.data_source == 'ob-raw-dump':
        dataset_info = parse_ob_raw_feature_information(args.data_path)

    elif args.data_source == 'ob-vw':
        dataset_info = parse_ob_vw_feature_information(args.data_path)

    elif args.data_source == 'ob-csv':
        dataset_info = parse_csv_with_description_information(args.data_path)

    elif args.data_source == 'csv-raw':
        dataset_info = parse_csv_raw(args.data_path)
    else:
        raise NotImplementedError(
            'Plase, select a supported data source. Possible sources: {csv-raw, ob-vw, ob-csv}',
        )

    return dataset_info


def display_tool_name() -> None:
    tool_name = """


                        *///////////////.
                     //////////////////////*
                   */////////////////////////.
                  ////////////// */////////////
                  /////////*          /////////
                 //////   /////   ////,   /////
                  ////////     ///    /////////
                  /////   /////  ./////   ////*
                   ,////                 ////
                     *////             ////.
                         ///////*///////


    ░█████╗░██╗░░░██╗████████╗██████╗░░█████╗░███╗░░██╗██╗░░██╗
    ██╔══██╗██║░░░██║╚══██╔══╝██╔══██╗██╔══██╗████╗░██║██║░██╔╝
    ██║░░██║██║░░░██║░░░██║░░░██████╔╝███████║██╔██╗██║█████═╝░
    ██║░░██║██║░░░██║░░░██║░░░██╔══██╗██╔══██║██║╚████║██╔═██╗░
    ╚█████╔╝╚██████╔╝░░░██║░░░██║░░██║██║░░██║██║░╚███║██║░╚██╗
    ░╚════╝░░╚═════╝░░░░╚═╝░░░╚═╝░░╚═╝╚═╝░░╚═╝╚═╝░░╚══╝╚═╝░░╚═╝


    """

    print(tool_name)


def parse_ob_line(
    line_string: str, delimiter: str = '\t', args: Any = None,
) -> list[str]:
    """Outbrain line parsing - generic TSVs"""

    line_string = line_string.strip()
    parts = line_string.split(delimiter)
    return parts


def parse_ob_line_vw(
    line_string: str,
    delimiter: str,
    args: Any = None,
    fw_col_mapping  = None,
    table_header  = None,
    include_namespace_info = False,
) -> list[str | None]:
    """Parse a sparse vw line into a pandas df with pre-defined namespace"""

    all_line_parts = line_string.strip().split('|')
    label_part = all_line_parts[0].split(' ')[0]
    remainder = all_line_parts[1:]
    label = label_part
    remainder_hash = dict()

    # Hash multi-value tuples and store name-val mappings
    for remaining_part in remainder:
        core_parts = remaining_part.split(' ')
        namespace_part = core_parts[0]
        other_parts = '-'.join(x for x in core_parts[1:] if x != '')
        if namespace_part in fw_col_mapping:
            remainder_hash[fw_col_mapping[namespace_part]] = other_parts

    # Construct the consistently-mapped instance based on the remainder mapping
    the_real_instance = [
        remainder_hash.get(
            el, None,
        ) for el in table_header[1:]
    ]
    if not include_namespace_info:
        the_real_instance = [
            x[2:] if not x is None else None for x in the_real_instance
        ]

    parts = [label] + the_real_instance
    return parts


def parse_ob_csv_line(
    line_string: str, delimiter: str = ',', args: Any = None,
) -> list[str]:
    """Data can have commas within JSON field dumps"""

    clx = list(csv.reader([line_string])).pop()
    return clx


def generic_line_parser(
    line_string: str,
    delimiter: str,
    args: Any = None,
    fw_col_mapping: Any = None,
    table_header: Any = None,
) -> list[Any]:
    """A generic method aimed to parse data from different sources."""

    if args.data_source == 'ob-raw-dump':
        return parse_ob_line(line_string, delimiter, args)

    elif args.data_source == 'ob-vw':
        return parse_ob_line_vw(
            line_string, delimiter, args, fw_col_mapping, table_header,
        )

    elif args.data_source == 'ob-csv' or args.data_source == 'csv-raw':
        return parse_ob_csv_line(line_string, delimiter, args)

    else:
        raise NotImplementedError(
            'Please, specify a valid --data_source argument!',
        )


def read_reference_json(json_path) -> dict[str, dict]:
    """A helper method for reading a JSON"""
    with open(json_path) as jp:
        return json.load(jp)


def parse_namespace(namespace_path: str) -> tuple[set[str], dict[str, str]]:
    """Parse the feature namespace for type awareness"""

    float_set = set()
    id_feature_map = {}

    with open(namespace_path) as nm:
        for line in nm:
            try:
                namespace_parts = line.strip().split(',')
                if len(namespace_parts) == 2 and '_' not in namespace_parts[0]:
                    fw_id, feature = namespace_parts
                    type_name = 'generic'

                else:
                    fw_id, feature, type_name = namespace_parts

                id_feature_map[fw_id] = feature
                if type_name == 'f32':
                    float_set.add(feature)
            except Exception as es:
                logging.error(f'\U0001F631 {es} -- {namespace_parts}')

    return float_set, id_feature_map


def read_column_names(mapping_file: str) -> list[str]:
    """Read the col. header"""

    with open(mapping_file, encoding='utf-8') as mf:
        columns = mf.read().strip().split('\t')
    return columns


def parse_ob_vw_feature_information(data_path) -> DatasetInformationStorage:
    """A generic parser of ob-based data"""

    # Get column names
    column_descriptions = os.path.join(data_path, 'vw_namespace_map.csv')
    column_types, fw_map = parse_namespace(column_descriptions)

    # We establish column order here
    column_names = ['label'] + list(fw_map.values())

    data_path = os.path.join(data_path, 'data.vw.gz')
    col_delimiter = None
    encoding = 'utf-8'

    return DatasetInformationStorage(
        data_path, column_names, column_types, col_delimiter, encoding, fw_map,
    )


def parse_ob_raw_feature_information(data_path) -> DatasetInformationStorage:
    """A generic parser of ob-based data"""

    # Get column names
    column_types: list[str] = []

    # Get set of numeric columns
    table_header_path = os.path.join(data_path, 'raw_data/0_header/header.csv')
    table_header = read_column_names(table_header_path)

    data_path_train = os.path.join(data_path, 'raw_data/1_train/*')
    col_delimiter = '\t'
    encoding = 'utf-8'

    final_df = []
    core_data_folders = glob.glob(data_path_train)
    for actual_data in core_data_folders:
        for dump in glob.glob(actual_data + '/*'):
            tmp_df = pd.read_csv(
                dump, sep='\t', low_memory=True, dtype='object',
            )
            assert tmp_df.shape[1] == len(table_header)
            tmp_df.columns = table_header
            final_df.append(tmp_df)

    final_df_concat = pd.concat(final_df, axis=0)
    final_path = os.path.join(data_path, 'raw_dump.tsv')
    logging.info(
        f'Stored data dump of dimension {final_df_concat.shape} to {final_path}',
    )
    final_df_concat.to_csv(final_path, sep='\t', index=False)
    data_path = os.path.join(data_path, 'raw_dump.tsv')

    return DatasetInformationStorage(
        data_path, table_header, set(column_types), col_delimiter, encoding, None,
    )


def parse_ob_feature_information(data_path) -> DatasetInformationStorage:
    """A generic parser of ob-based data"""

    # Get column names
    column_names = os.path.join(data_path, 'vw_namespace_map.csv')
    column_types, _ = parse_namespace(column_names)

    # Get set of numeric columns
    table_header_path = os.path.join(data_path, 'raw_data/0_header/header.csv')
    table_header = read_column_names(table_header_path)

    data_path = os.path.join(data_path, 'raw_data/1_train/*')
    col_delimiter = '\t'
    encoding = 'utf-8'

    return DatasetInformationStorage(
        data_path, table_header, column_types, col_delimiter, encoding, None,
    )


def parse_csv_with_description_information(data_path) -> DatasetInformationStorage:
    dataset_description = read_reference_json(
        os.path.join(data_path, 'dataset_desc.json'),
    )
    column_names = []
    column_types = set()
    for feature in dataset_description.get('data_features', []):
        feature_name = feature.get('name')
        column_names.append(feature_name)
        feature_type = feature.get('type', '')
        if 'float' in feature_type or 'Float' in feature_type:
            column_types.add(feature_name)
    col_delimiter = ','
    data_path = os.path.join(data_path, 'data.csv')
    encoding = 'latin1'
    return DatasetInformationStorage(
        data_path, column_names, column_types, col_delimiter, encoding, None,
    )


def parse_csv_raw(data_path) -> DatasetInformationStorage:
    column_types: set[str] = set()

    data_path = os.path.join(data_path, 'data.csv')
    with open(data_path) as inp_data:
        header = inp_data.readline()
    col_delimiter = ','
    column_names = header.strip().split(col_delimiter)
    encoding = 'latin1'
    return DatasetInformationStorage(
        data_path, column_names, column_types, col_delimiter, encoding, None,
    )


def extract_features_from_reference_JSON(json_path: str) -> set[Any]:
    """Given a model's JSON, extract unique features"""

    with open(json_path) as jp:
        content = json.load(jp)

    unique_features = set()
    feature_space = content['desc'].get('features', [])
    fields_space = content['desc'].get('fields', [])
    joint_space = feature_space + fields_space

    for feature_tuple in joint_space:
        for individual_feature in feature_tuple.split(','):
            unique_features.add(individual_feature)

    return unique_features


def summarize_feature_bounds_for_transformers(
    bounds_object_storage: Any,
    feature_types: list[str],
    task_name: str,
    label_name: str,
    granularity: int = 15,
    output_summary_table_only: bool = False,
):
    """summarization auxilliary method for generating JSON-based specs"""

    if bounds_object_storage is None:
        logging.info('Bounds storage object is empty.')
        exit()

    final_storage = defaultdict(list)
    for el in bounds_object_storage:
        if isinstance(el, dict):
            for k, v in el.items():
                final_storage[k].append(v)

    summary_table_rows = []
    for k, v in final_storage.items():
        # Conduct local aggregation + bound changes
        if k in feature_types and k != label_name:
            minima, maxima, medians, uniques = [], [], [], []
            for feature_summary in v:
                minima.append(feature_summary.minimum)
                maxima.append(feature_summary.maximum)
                medians.append(feature_summary.median)
                uniques.append(feature_summary.num_unique)
            summary_table_rows.append(
                [
                    k,
                    round(np.min(minima), 2),
                    round(np.max(maxima), 2),
                    round(np.median(medians), 2),
                    int(np.mean(uniques)),
                ],
            )

    if len(summary_table_rows) == 0:
        logging.info('No numeric features to summarize.')
        return None

    summary_table: pd.Dataframe = pd.DataFrame(summary_table_rows)
    summary_table.columns = [
        'Feature',
        'Minimum',
        'Maximum',
        'Median',
        'Num avg. unique (batch)',
    ]

    if output_summary_table_only:
        return summary_table

    if len(summary_table) == 0:
        logging.info('Summary table empty, skipping transformer generation ..')
        return

    if task_name == 'feature_summary_transformers':
        transformers_per_feature = defaultdict(list)

        # Take care of weights first -> range is pre-defined
        for k, v in final_storage.items():
            if label_name in k or 'dummy' in k:
                continue

            weight_template = {
                'feature': k,
                'src_features': [k],
                'transformations': ['Weight'],
                'weights': [0, 0.5, 1.5, 2, 3, 10],
            }
            transformers_per_feature[k].append(weight_template)

        # Consider numeric transformations - pairs and single ones
        for enx, row in summary_table.iterrows():
            if row.Feature == 'dummy':
                continue
            try:
                actual_range = (
                    np.arange(
                        row['Minimum'],
                        row['Maximum'],
                        (row['Maximum'] - row['Minimum']) / granularity,
                    )
                    .round(2)
                    .tolist()
                )
                binner_template = {
                    'feature': f'{row.Feature}',
                    'src_features': [row.Feature],
                    'transformations': [
                        'BinnerSqrt',
                        'BinnerLog',
                        'BinnerSqrtPlain',
                        'BinnerLogPlain',
                    ],
                    'n': actual_range,
                    'resolutions': [0.1, 2, 4, 8, 16, 32, 64, 128],
                }

            except Exception as es:
                logging.info(
                    f'\U0001F631 Encountered {es}. The problematic feature is: {row}, skipping transformer for this feature ..',
                )

            transformers_per_feature[row.Feature].append(binner_template)

            # We want the full loop here, due to asymmetry of transformation(s)
            for enx_second, row_second in summary_table.iterrows():
                if enx_second < enx:
                    continue

                # The n values are defined based on maxima of the second feature
                if row_second.Feature != row.Feature:
                    n_bound = round(row_second['Median'] + row['Median'], 2)
                    max_bound = round(
                        min(row_second['Maximum'], row['Maximum']), 2,
                    )
                    min_bound = round(
                        row_second['Minimum'] + row['Minimum'], 2,
                    )
                    range_spectrum = sorted(
                        list(
                            {
                                0.0,
                                min_bound,
                                n_bound / 10,
                                n_bound / 5,
                                n_bound,
                                max_bound,
                            },
                        ),
                    )

                    range_spectrum = [x for x in range_spectrum if x >= 0]
                    binner_pair_template = {
                        'feature': f'{row.Feature}Ratio{row_second.Feature}',
                        'src_features': [row.Feature, row_second.Feature],
                        'transformations': ['BinnerLogRatioPlain', 'BinnerLogRatio'],
                        'n': range_spectrum,
                        'resolutions': [0.1, 2, 4, 8, 16, 32, 64, 128],
                    }

                    binner_pair_template_second = {
                        'feature': f'{row_second.Feature}Ratio{row.Feature}',
                        'src_features': [row_second.Feature, row.Feature],
                        'transformations': ['BinnerLogRatioPlain', 'BinnerLogRatio'],
                        'n': range_spectrum,
                        'resolutions': [0.1, 2, 4, 8, 16, 32, 64, 128],
                    }

                    transformers_per_feature[row.Feature].append(
                        binner_pair_template,
                    )
                    transformers_per_feature[row.Feature].append(
                        binner_pair_template_second,
                    )

        binner_templates = []
        for k, v in transformers_per_feature.items():
            for transformer_struct in v:
                binner_templates.append(transformer_struct)

        logging.info(
            f'Generated {len(binner_templates)} transformation search specifications.\n',
        )
        namespace_full = f'"random_grid_feature_transform": {json.dumps(binner_templates)}, "random_grid_epochs": 512'
        logging.info('Generated transformations below:\n')
        print(namespace_full)


def summarize_rare_counts(
    term_counter: Any,
    args: Any,
    cardinality_object: Any,
    object_info: DatasetInformationStorage,
) -> None:
    """Write rare values"""

    out_df_rows = []
    logging.info(
        f'Rare value summary (freq <= {args.rare_value_count_upper_bound}) follows ..',
    )

    for namespace_tuple, count in term_counter.items():
        namespace, value = namespace_tuple
        out_df_rows.append([namespace, value, count])
    out_df: pd.DataFrame = pd.DataFrame(out_df_rows)
    out_df.columns = ['Namespace', 'value', 'Count']
    out_df.to_csv(
        os.path.join(args.output_folder, 'rare_values.tsv'), sep='\t', index=False,
    )
    logging.info(f'Wrote rare values to {args.output_folder}/rare_values.tsv')

    overall_rare_counts = Counter(out_df.Namespace.values)
    sorted_counts = sorted(
        overall_rare_counts.items(), key=lambda pair: pair[1], reverse=True,
    )
    for k, v in sorted_counts:
        logging.info(f'Namespace: {k} ---- Rare values observed: {v}')

    final_df_rows = []
    for k, v in sorted_counts:
        cardinality = len(cardinality_object[k])
        rare_proportion = np.round(100 * (v / cardinality), 2)
        col_type = 'nominal'
        if k in object_info.column_types:
            col_type = 'numeric'
        final_df_rows.append(
            {
                'rare_proportion': rare_proportion,
                'feature_type': col_type,
                'feature_name': k,
            },
        )

    final_df: pd.DataFrame = pd.DataFrame(final_df_rows)
    final_df = final_df.sort_values(by=['rare_proportion'])
    logging.info(
        f'Wrote feature sparsity summary to {args.output_folder}/feature_sparsity_summary.tsv',
    )
    final_df.to_csv(
        f'{args.output_folder}/feature_sparsity_summary.tsv', index=False, sep='\t',
    )



from __future__ import annotations

import gzip
import itertools
import logging
import os
import random
import time
from collections import Counter
from collections import defaultdict
from collections import deque
from timeit import default_timer as timer
from typing import Any
from typing import Dict
from typing import List
from typing import Set
from typing import Tuple
from typing import Union

import numpy as np
import pandas as pd
import tqdm

from outrank.algorithms.importance_estimator import get_importances_estimate_pairwise
from outrank.algorithms.sketches.counting_ultiloglog import (
    HyperLogLogWCache as HyperLogLog,
)
from outrank.core_utils import BatchRankingSummary
from outrank.core_utils import extract_features_from_reference_JSON
from outrank.core_utils import generic_line_parser
from outrank.core_utils import internal_hash
from outrank.core_utils import NominalFeatureSummary
from outrank.core_utils import NumericFeatureSummary
from outrank.feature_transformations.ranking_transformers import FeatureTransformerGeneric
from outrank.feature_transformations.ranking_transformers import FeatureTransformerNoise

logger = logging.getLogger('syn-logger')
logger.setLevel(logging.DEBUG)
random.seed(a=123, version=2)
GLOBAL_CARDINALITY_STORAGE: dict[Any, Any] = dict()
GLOBAL_RARE_VALUE_STORAGE: dict[str, Any] = Counter()

IGNORED_VALUES = set()
HYPERLL_ERROR_BOUND = 0.02


def encode_int_column(input_tuple: tuple[str, Any]) -> tuple[Any, list[int]]:
    """Encode column values as categoric (at a batch level!)"""

    hashes, _ = pd.factorize(input_tuple[1])
    return input_tuple[0], hashes


def mixed_rank_graph(
    input_dataframe: pd.DataFrame, args: Any, cpu_pool: Any, pbar: Any,
) -> BatchRankingSummary:
    """Compute the full mixed rank graph corresponding to all pairwise feature interactions based on the selected heuristic"""

    all_columns = input_dataframe.columns

    triplets = []
    tmp_df = input_dataframe.copy()
    out_time_struct = {}

    # Handle cont. types prior to interaction evaluation
    pbar.set_description('Encoding columns')
    jobs = [(cname, tmp_df[cname]) for cname in all_columns]
    col_dots = '.'
    start_enc_timer = timer()
    with cpu_pool as p:
        results = p.amap(encode_int_column, jobs)
        while not results.ready():
            time.sleep(4)
            col_dots = col_dots + '.'
            pbar.set_description(f'Encoding columns .{col_dots}')
        tmp_df = pd.DataFrame({k: v for k, v in results.get()})
    end_enc_timer = timer()
    out_time_struct['encoding_columns'] = end_enc_timer - start_enc_timer

    # Helper method for parallel estimation
    combinations = list(
        itertools.combinations_with_replacement(all_columns, 2),
    )

    if '3mr' in args.heuristic:
        rel_columns = [
            column for column in all_columns if ' AND_REL ' in column
        ]
        non_rel_columns = list(set(all_columns) - set(rel_columns))
        combinations = list(
            itertools.combinations_with_replacement(non_rel_columns, 2),
        )
        combinations += [(column, args.label_column) for column in rel_columns]
    else:
        combinations = list(
            itertools.combinations_with_replacement(all_columns, 2),
        )

    # Diagonal elements
    for individual_column in all_columns:
        if individual_column != args.label_column:
            combinations += [(individual_column, individual_column)]

    # Some applications do not require the full feature-feature triangular matrix
    if (args.target_ranking_only == 'True') and ('3mr' not in args.heuristic):
        combinations = [x for x in combinations if args.label_column in x]

    random.shuffle(combinations)
    combinations = combinations[: args.combination_number_upper_bound]

    if args.heuristic == 'Constant':
        final_constant_imp = []
        for c1, c2 in combinations:
            final_constant_imp.append((c1, c2, 0.0))

        out_time_struct['feature_score_computation'] = end_enc_timer - \
            start_enc_timer
        return BatchRankingSummary(final_constant_imp, out_time_struct)

    # Map the scoring calls to the worker pool
    pbar.set_description('Allocating thread pool')

    # starmap is an alternative that is slower unfortunately (but nicer)
    def get_grounded_importances_estimate(combination: tuple[str]) -> Any:
        return get_importances_estimate_pairwise(combination, args, tmp_df=tmp_df)

    start_enc_timer = timer()
    with cpu_pool as p:
        pbar.set_description(f'Computing (#ftr={len(combinations)})')
        results = p.amap(get_grounded_importances_estimate, combinations)
        while not results.ready():
            time.sleep(4)
        triplets = results.get()
    end_enc_timer = timer()
    out_time_struct['feature_score_computation'] = end_enc_timer - \
        start_enc_timer

    # Gather the final triplets
    pbar.set_description('Aggregation of ranking results')
    final_triplets = []
    for triplet in triplets:
        inv = (triplet[1], triplet[0], triplet[2])
        final_triplets.append(inv)
        final_triplets.append(triplet)
        triplets = final_triplets

    pbar.set_description('Proceeding to the next batch of data')
    return BatchRankingSummary(triplets, out_time_struct)


def enrich_with_transformations(
    input_dataframe: pd.DataFrame, num_col_types: set[str], logger: Any, args: Any,
) -> pd.DataFrame:
    """Construct a collection of new features based on pre-defined transformations/rules"""

    transformer = FeatureTransformerGeneric(
        num_col_types, preset=args.transformers,
    )
    transformed_df = transformer.construct_new_features(input_dataframe)
    logger.info(
        f'Constructed {len(transformer.constructed_feature_names)} new features ..',
    )

    return transformed_df


def compute_combined_features(
    input_dataframe: pd.DataFrame,
    logger: Any,
    args: Any,
    pbar: Any,
    is_3mr: bool = False,
) -> pd.DataFrame:
    """Compute higher order features via xxhash-based trick."""

    all_columns = [
        x for x in input_dataframe.columns if x != args.label_column
    ]
    join_string = ' AND_REL ' if is_3mr else ' AND '
    interaction_order = 2 if is_3mr else args.interaction_order

    full_combination_space = list(
        itertools.combinations(all_columns, interaction_order),
    )

    if args.combination_number_upper_bound:
        random.shuffle(full_combination_space)
        full_combination_space = full_combination_space[
            : args.combination_number_upper_bound
        ]

    com_counter = 0
    new_feature_hash = {}
    for new_combination in full_combination_space:
        pbar.set_description(
            f'Created {com_counter}/{len(full_combination_space)}',
        )
        combined_feature: list[str] = [str(0)] * input_dataframe.shape[0]
        for feature in new_combination:
            tmp_feature = input_dataframe[feature].tolist()
            for enx, el in enumerate(tmp_feature):
                combined_feature[enx] = str(
                    internal_hash(
                        str(combined_feature[enx]) + str(el),
                    ),
                )
        ftr_name = join_string.join(str(x) for x in new_combination)
        new_feature_hash[ftr_name] = combined_feature
        com_counter += 1
    tmp_df = pd.DataFrame(new_feature_hash)
    pbar.set_description('Concatenating into final frame ..')
    input_dataframe = pd.concat([input_dataframe, tmp_df], axis=1)
    del tmp_df

    return input_dataframe


def compute_expanded_multivalue_features(
    input_dataframe: pd.DataFrame, logger: Any, args: Any, pbar: Any,
) -> pd.DataFrame:
    """Compute one-hot encoded feature space based on each designated multivalue feature. E.g., feature with value "a,b,c" becomes three features, values of which are presence of a given value in a mutlivalue feature of choice."""

    considered_multivalue_features = args.explode_multivalue_features.split(
        ';',
    )
    new_feature_hash = {}
    missing_symbols = set(args.missing_value_symbols.split(','))

    for multivalue_feature in considered_multivalue_features:
        multivalue_feature_vector = input_dataframe[multivalue_feature].values.tolist(
        )
        multivalue_feature_vector = [
            x.replace(',', '-') for x in multivalue_feature_vector
        ]
        multivalue_sets = [
            set(x.split('-'))
            for x in multivalue_feature_vector
        ]
        unique_values = set.union(*multivalue_sets)

        for missing_symbol in missing_symbols:
            if missing_symbol in unique_values:
                unique_values.remove(missing_symbol)

        for unique_value in unique_values:
            tmp_vec = []
            for enx, multivalue in enumerate(multivalue_sets):
                if unique_value in multivalue:
                    tmp_vec.append('1')
                else:
                    tmp_vec.append('')

            new_feature_hash[f'MULTIEX-{multivalue_feature}-{unique_value}'] = tmp_vec

    tmp_df = pd.DataFrame(new_feature_hash)
    input_dataframe = pd.concat([input_dataframe, tmp_df], axis=1)
    del tmp_df

    return input_dataframe


def compute_subfeatures(
    input_dataframe: pd.DataFrame, logger: Any, args: Any, pbar: Any,
) -> pd.DataFrame:
    """Compute derived features that are more fine-grained. Implements logic around two operators that govern feature construction.
    ->: One sided construction - every value from left side is fine, separate ones from the right side feature will be considered.
    <->: Two sided construction - two-sided values present. This means that each value from a is combined with each from b, forming |A|*|B| new features (one-hot encoded)
    """

    all_subfeature_pair_seeds = args.subfeature_mapping.split(';')
    new_feature_hash = dict()

    for seed_pair in all_subfeature_pair_seeds:
        if '<->' in seed_pair:
            feature_first, feature_second = seed_pair.split('<->')

        elif '->' in seed_pair:
            feature_first, feature_second = seed_pair.split('->')

        else:
            raise NotImplementedError(
                'Please specify valid subfeature operator (<-> or ->)',
            )

        subframe = input_dataframe[[feature_first, feature_second]]
        unique_feature_second = subframe[feature_second].unique()
        feature_first_vec = subframe[feature_first].tolist()
        feature_second_vec = subframe[feature_second].tolist()
        out_template_feature = [
            (a, b) for a, b in zip(feature_first_vec, feature_second_vec)
        ]

        if '<->' in seed_pair:
            unique_feature_first = subframe[feature_first].unique()

            mask_types = []
            for unique_target_feature_value in unique_feature_second:
                for unique_seed_feature_value in unique_feature_first:
                    mask_types.append(
                        (unique_seed_feature_value, unique_target_feature_value),
                    )

            for mask_type in mask_types:
                new_feature = []
                for value_tuple in out_template_feature:
                    if (
                        value_tuple[0] == mask_type[0]
                        and value_tuple[1] == mask_type[1]
                    ):
                        new_feature.append(str(1))
                    else:
                        new_feature.append(str(0))
                feature_name = (
                    f'SUBFEATURE|{feature_first}|{feature_second}-'
                    + mask_type[0]
                    + '&'
                    + mask_type[1]
                )
                new_feature_hash[feature_name] = new_feature

            del new_feature

        elif '->' in seed_pair:
            for unique_target_feature_value in unique_feature_second:
                tmp_new_feature = [
                    'AND'.join(
                        x,
                    ) if x[1] == unique_target_feature_value else ''
                    for x in out_template_feature
                ]
                feature_name_final = (
                    'SUBFEATURE-' + feature_first + '&' + unique_target_feature_value
                )
                new_feature_hash[feature_name_final] = tmp_new_feature

    tmp_df = pd.DataFrame(new_feature_hash)
    input_dataframe = pd.concat([input_dataframe, tmp_df], axis=1)

    del tmp_df
    return input_dataframe


def include_noisy_features(
    input_dataframe: pd.DataFrame, logger: Any, args: Any,
) -> pd.DataFrame:
    """Add randomized features that serve as a sanity check"""

    transformer = FeatureTransformerNoise()
    transformed_df = transformer.construct_new_features(
        input_dataframe, args.label_column,
    )

    return transformed_df


def compute_coverage(input_dataframe: pd.DataFrame, args: Any) -> dict[str, set[str]]:
    """Compute coverage of features, incrementally"""
    output_storage_cov = defaultdict(set)
    all_missing_symbols = set(args.missing_value_symbols.split(','))
    for column in input_dataframe:
        all_missing = sum(
            [
                input_dataframe[column].values.tolist().count(x)
                for x in all_missing_symbols
            ],
        )

        output_storage_cov[column] = (
            1 - (all_missing / input_dataframe.shape[0])
        ) * 100

    return output_storage_cov


def compute_feature_memory_consumption(input_dataframe: pd.DataFrame, args: Any) -> dict[str, set[str]]:
    """An approximation of how much feature take up"""
    output_storage_features = defaultdict(set)
    for col in input_dataframe.columns:
        specific_column = [
            str(x).strip() for x in input_dataframe[col].astype(str).values.tolist()
        ]
        col_size = sum(
            len(x.encode())
            for x in specific_column
        ) / input_dataframe.shape[0]
        output_storage_features[col] = col_size
    return output_storage_features


def compute_value_counts(input_dataframe: pd.DataFrame, args: Any):
    """Update the count structure"""

    global GLOBAL_RARE_VALUE_STORAGE
    global IGNORED_VALUES

    for column in input_dataframe.columns:
        main_values = input_dataframe[column].values
        for value in main_values:
            if value not in IGNORED_VALUES:
                GLOBAL_RARE_VALUE_STORAGE.update({(column, value): 1})

    for key, val in GLOBAL_RARE_VALUE_STORAGE.items():
        if val > args.rare_value_count_upper_bound:
            IGNORED_VALUES.add(key)

    for to_remove_val in IGNORED_VALUES:
        del GLOBAL_RARE_VALUE_STORAGE[to_remove_val]


def compute_cardinalities(input_dataframe: pd.DataFrame, pbar: Any) -> None:
    """Compute cardinalities of features, incrementally"""

    global GLOBAL_CARDINALITY_STORAGE
    output_storage_card = defaultdict(set)
    for enx, column in enumerate(input_dataframe):
        output_storage_card[column] = set(input_dataframe[column].unique())
        if column not in GLOBAL_CARDINALITY_STORAGE:
            GLOBAL_CARDINALITY_STORAGE[column] = HyperLogLog(
                HYPERLL_ERROR_BOUND,
            )

        for unique_value in set(input_dataframe[column].unique()):
            if unique_value:
                GLOBAL_CARDINALITY_STORAGE[column].add(
                    internal_hash(unique_value),
                )
        pbar.set_description(
            f'Computing cardinality (Hyperloglog update) {enx}/{input_dataframe.shape[1]}',
        )


def compute_bounds_increment(
    input_dataframe: pd.DataFrame, numeric_column_types: set[str],
) -> dict[str, Any]:
    all_features = input_dataframe.columns
    numeric_column_types = set(numeric_column_types)
    summary_object = {}
    summary_storage: Any = {}
    for feature in all_features:
        if feature in numeric_column_types:
            feature_vector = pd.to_numeric(
                input_dataframe[feature], errors='coerce',
            )
            minimum = np.min(feature_vector)
            maximum = np.max(feature_vector)
            mean = np.mean(feature_vector)
            summary_storage = NumericFeatureSummary(
                feature, minimum, maximum, mean, len(
                    np.unique(feature_vector),
                ),
            )
            summary_object[feature] = summary_storage

        else:
            feature_vector = input_dataframe[feature].values
            summary_storage = NominalFeatureSummary(
                feature, len(np.unique(feature_vector)),
            )
            summary_object[feature] = summary_storage

    return summary_object


def compute_batch_ranking(
    line_tmp_storage: list[list[Any]],
    numeric_column_types: set[str],
    args: Any,
    cpu_pool: Any,
    column_descriptions: list[str],
    logger: Any,
    pbar: Any,
) -> tuple[BatchRankingSummary, dict[str, Any], dict[str, set[str]], dict[str, set[str]]]:
    """Enrich the feature space and compute the batch importances"""

    input_dataframe = pd.DataFrame(line_tmp_storage)
    input_dataframe.columns = column_descriptions
    pbar.set_description('Control features')

    if args.feature_set_focus:
        if args.feature_set_focus == '_all_from_reference_JSON':
            focus_set = extract_features_from_reference_JSON(
                args.reference_model_JSON,
            )

        else:
            focus_set = set(args.feature_set_focus.split(','))

        focus_set.add(args.label_column)
        focus_set = {x for x in focus_set if x in input_dataframe.columns}
        input_dataframe = input_dataframe[focus_set]

    if args.transformers != 'none':
        pbar.set_description('Adding transformations')
        input_dataframe = enrich_with_transformations(
            input_dataframe, numeric_column_types, logger, args,
        )

    if args.explode_multivalue_features != 'False':
        pbar.set_description('Constructing new features from multivalue ones')
        input_dataframe = compute_expanded_multivalue_features(
            input_dataframe, logger, args, pbar,
        )

    if args.subfeature_mapping != 'False':
        pbar.set_description('Constructing new (sub)features')
        input_dataframe = compute_subfeatures(
            input_dataframe, logger, args, pbar,
        )

    if args.interaction_order > 1:
        pbar.set_description('Constructing new features')
        input_dataframe = compute_combined_features(
            input_dataframe, logger, args, pbar,
        )

    # in case of 3mr we compute the score of combinations against the target
    if '3mr' in args.heuristic:
        pbar.set_description(
            'Constructing features for computing relations in 3mr',
        )
        input_dataframe = compute_combined_features(
            input_dataframe, logger, args, pbar, True,
        )

    if args.include_noise_baseline_features == 'True' and args.heuristic != 'Constant':
        pbar.set_description('Computing baseline features')
        input_dataframe = include_noisy_features(input_dataframe, logger, args)

    # Compute incremental statistic useful for data inspection/transformer generation
    pbar.set_description('Computing coverage')
    coverage_storage = compute_coverage(input_dataframe, args)
    feature_memory_consumption = compute_feature_memory_consumption(
        input_dataframe, args,
    )
    compute_cardinalities(input_dataframe, pbar)

    if args.task == 'identify_rare_values':
        compute_value_counts(input_dataframe, args)

    bounds_storage = compute_bounds_increment(
        input_dataframe, numeric_column_types,
    )

    pbar.set_description(
        f'Computing ranks for {input_dataframe.shape[1]} features',
    )

    return (
        mixed_rank_graph(input_dataframe, args, cpu_pool, pbar),
        bounds_storage,
        coverage_storage,
        feature_memory_consumption,
    )


def get_num_of_instances(fname: str) -> int:
    """Count the number of lines in a file, fast - useful for progress logging"""

    def _make_gen(reader):
        while True:
            b = reader(2**16)
            if not b:
                break
            yield b

    with open(fname, 'rb') as f:
        count = sum(buf.count(b'\n') for buf in _make_gen(f.raw.read))
    return count


def get_grouped_df(importances_df_list: list[tuple[str, str, float]]) -> pd.DataFrame:
    """A helper method that enables median-based aggregation after processing"""

    importances_df = pd.DataFrame(importances_df_list)
    if len(importances_df) == 0:
        return None
    importances_df.columns = ['FeatureA', 'FeatureB', 'Score']
    grouped = importances_df.groupby(
        ['FeatureA', 'FeatureB'],
    ).median().reset_index()
    return grouped


def checkpoint_importances_df(importances_batch: list[tuple[str, str, float]]) -> None:
    """A helper which stores intermediary state - useful for longer runs"""

    gdf = get_grouped_df(importances_batch)
    if gdf is not None:
        gdf.to_csv('ranking_checkpoint_tmp.tsv', sep='\t')


def estimate_importances_minibatches(
    input_file: str,
    column_descriptions: list,
    fw_col_mapping: dict[str, str],
    numeric_column_types: set,
    batch_size: int = 100000,
    args: Any = None,
    data_encoding: str = 'utf-8',
    cpu_pool: Any = None,
    delimiter: str = '\t',
    feature_construction_mode: bool = False,
    logger: Any = None,
) -> tuple[list[dict[str, Any]], Any, dict[Any, Any], list[dict[str, Any]], list[dict[str, set[str]]], defaultdict[str, list[set[str]]], dict[str, Any]]:
    """Interaction score estimator - suitable for example for csv-like input data types.
    This type of data is normally a single large csv, meaning that minibatch processing needs to
    happen during incremental handling of the file (that"s not the case for pre-separated ob data)
    """

    invalid_line_queue: Any = deque([], maxlen=2**5)

    invalid_lines = 0
    line_counter = 0

    importances_df: list[Any] = []
    line_tmp_storage = []
    bounds_storage_batch = []
    memory_storage_batch = []
    step_timing_checkpoints = []

    local_coverage_object = defaultdict(list)
    local_pbar = tqdm.tqdm(
        total=get_num_of_instances(input_file) - 1, position=0,
    )

    file_name, file_extension = os.path.splitext(input_file)

    if file_extension == '.gz':
        file_stream = gzip.open(input_file, 'rt', encoding=data_encoding)

    else:
        file_stream = open(input_file, encoding=data_encoding)

    file_stream.readline()

    local_pbar.set_description('Starting ranking computation')
    for line in file_stream:
        line_counter += 1
        local_pbar.update(1)

        if line_counter % args.subsampling != 0:
            continue

        parsed_line = generic_line_parser(
            line, delimiter, args, fw_col_mapping, column_descriptions,
        )

        if len(parsed_line) == len(column_descriptions):
            line_tmp_storage.append(parsed_line)

        else:
            invalid_line_queue.appendleft(str(parsed_line))
            invalid_lines += 1

        # Batches need to be processed on-the-fly
        if len(line_tmp_storage) >= args.minibatch_size:

            importances_batch, bounds_storage, coverage_storage, memory_storage = compute_batch_ranking(
                line_tmp_storage,
                numeric_column_types,
                args,
                cpu_pool,
                column_descriptions,
                logger,
                local_pbar,
            )

            bounds_storage_batch.append(bounds_storage)
            memory_storage_batch.append(memory_storage)
            for k, v in coverage_storage.items():
                local_coverage_object[k].append(v)

            del coverage_storage

            line_tmp_storage = []
            step_timing_checkpoints.append(importances_batch.step_times)
            importances_df += importances_batch.triplet_scores

            if args.heuristic != 'Constant':
                local_pbar.set_description('Creating checkpoint')
                checkpoint_importances_df(importances_df)

    file_stream.close()

    local_pbar.set_description('Parsing the remainder')
    if invalid_lines > 0:
        logger.info(
            f"Detected {invalid_lines} invalid lines. If this number is very high, it's possible your header is off - re-check your data/attribute-feature mappings please!",
        )

        invalid_lines_log = '\n INVALID_LINE ====> '.join(
            list(invalid_line_queue)[0:5],
        )
        logger.info(
            f'5 samples of invalid lines are printed below\n {invalid_lines_log}',
        )

    remaining_batch_size = len(line_tmp_storage)

    if remaining_batch_size > 2**10:
        line_tmp_storage = line_tmp_storage[: args.minibatch_size]
        importances_batch, bounds_storage, coverage_storage, _ = compute_batch_ranking(
            line_tmp_storage,
            numeric_column_types,
            args,
            cpu_pool,
            column_descriptions,
            logger,
            local_pbar,
        )

        for k, v in coverage_storage.items():
            local_coverage_object[k].append(v)

        step_timing_checkpoints.append(importances_batch.step_times)
        importances_df += importances_batch.triplet_scores
        bounds_storage = dict()
        bounds_storage_batch.append(bounds_storage)
        checkpoint_importances_df(importances_df)

    local_pbar.set_description('Wrapping up')
    local_pbar.close()

    return (
        step_timing_checkpoints,
        get_grouped_df(importances_df),
        GLOBAL_CARDINALITY_STORAGE,
        bounds_storage_batch,
        memory_storage_batch,
        local_coverage_object,
        GLOBAL_RARE_VALUE_STORAGE,
    )



from __future__ import annotations

import glob
import logging
import os
import signal
from typing import Any

import numpy as np
import pandas as pd

from outrank.algorithms.importance_estimator import rank_features_3MR
from outrank.core_ranking import estimate_importances_minibatches
from outrank.core_utils import display_random_tip
from outrank.core_utils import display_tool_name
from outrank.core_utils import get_dataset_info
from outrank.core_utils import summarize_feature_bounds_for_transformers
from outrank.core_utils import summarize_rare_counts
from outrank.core_utils import write_json_dump_to_file

logging.basicConfig(format='%(asctime)s %(message)s', level=logging.INFO)
signal.signal(signal.SIGINT, signal.default_int_handler)

try:
    # pathos enables proper pickling during parallelization (multiprocessing does not)
    from pathos.multiprocessing import ProcessingPool as Pool

except Exception as es:
    logging.info(
        f'\U0001F631 Please install the "pathos" library (pip install pathos) for required multithreading capabilities. {es}',
    )


def outrank_task_conduct_ranking(args: Any):
    # Data source = folder structure + relevant file specifications

    # No need for full-blown ranking in this case
    if args.task in ['identify_rare_values', 'feature_summary_transformers']:
        args.heuristic = 'Constant'

    display_tool_name()
    display_random_tip()

    dataset_info = get_dataset_info(args)

    for arg in vars(args):
        logging.info(f'{arg} set to: {getattr(args, arg)}')

    # Generate output folders (if not present)
    output_dir = os.path.dirname(
        os.path.join(
            args.output_folder, 'pairwise_ranks.tsv',
        ),
    )
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    # Initialize the global pool
    GLOBAL_CPU_POOL = Pool(args.num_threads)
    global_mutual_information_estimates = []
    global_bounds_storage = []
    global_memory_storage = []
    all_timings = []
    # Traverse the batches
    for raw_dump in glob.glob(dataset_info.data_path):

        if (
            args.data_source == 'ob-vw'
            or args.data_source == 'ob-csv'
            or args.data_source == 'csv-raw'
            or args.data_source == 'ob-raw-dump'
        ):
            all_subfiles = [raw_dump]

        for partial_data in all_subfiles:
            cmd_arguments = {
                'input_file': partial_data,
                'fw_col_mapping': dataset_info.fw_map,
                'column_descriptions': dataset_info.column_names,
                'numeric_column_types': dataset_info.column_types,
                'args': args,
                'data_encoding': dataset_info.encoding,
                'cpu_pool': GLOBAL_CPU_POOL,
                'delimiter': dataset_info.col_delimiter,
                'logger': logging,
            }

            if (
                args.data_source == 'ob-csv'
                or args.data_source == 'ob-vw'
                or args.data_source == 'csv-raw'
                or args.data_source == 'ob-raw-dump'
            ):
                (
                    checkpoint_timings,
                    mutual_information_estimates,
                    cardinality_object,
                    bounds_object_storage,
                    memory_object_storage,
                    coverage_object,
                    RARE_VALUE_STORAGE,
                ) = estimate_importances_minibatches(**cmd_arguments)

            global_bounds_storage += bounds_object_storage
            global_memory_storage += memory_object_storage
            all_timings += checkpoint_timings

            if cardinality_object is None:
                continue

            if coverage_object is None:
                continue

            if mutual_information_estimates is not None:
                global_mutual_information_estimates.append(
                    mutual_information_estimates,
                )

    if args.task == 'identify_rare_values':
        logging.info('Summarizing rare values ..')
        summarize_rare_counts(
            RARE_VALUE_STORAGE, args, cardinality_object, dataset_info,
        )
        exit()

    if args.task == 'feature_summary_transformers':
        summarize_feature_bounds_for_transformers(
            bounds_object_storage,
            dataset_info.column_types,
            args.task,
            args.label_column,
        )
        exit()
    else:
        summary_of_numeric_features = summarize_feature_bounds_for_transformers(
            bounds_object_storage,
            dataset_info.column_types,
            args.task,
            args.label_column,
            output_summary_table_only=True,
        )
        if summary_of_numeric_features is not None:
            num_out = os.path.join(
                args.output_folder, 'numeric_feature_statistics.tsv',
            )
            summary_of_numeric_features.to_csv(num_out, sep='\t', index=False)
            logging.info(
                f'Stored statistics of numeric features to {num_out} ..',
            )

    # Just in case.
    GLOBAL_CPU_POOL.close()
    GLOBAL_CPU_POOL.join()

    if len(global_mutual_information_estimates) == 0:
        logging.info('No rankings were obtained, exiting ..')
        exit()

    # Compute median imps across batches
    triplets = pd.concat(global_mutual_information_estimates, axis=0)
    triplets.columns = ['FeatureA', 'FeatureB', 'Score']

    if '3mr' in args.heuristic:
        # relevance include MI-scores of features w.r.t. labels
        relevance_df = triplets[triplets.FeatureB == args.label_column].copy()
        relevance_df = relevance_df[
            relevance_df.FeatureA.map(lambda x: ' AND_REL ' not in x)
        ][['FeatureA', 'Score']]
        relevance_df = relevance_df[relevance_df.FeatureA != args.label_column]

        # relations include MI-scores of combinations w.r.t. label
        relations_df = triplets[triplets.FeatureB == args.label_column][
            ['FeatureA', 'Score']
        ].copy()
        relations_df = relations_df[
            relations_df.FeatureA.map(lambda x: ' AND_REL ' in x)
        ]
        relations_df['FeatureB'] = relations_df.FeatureA.map(
            lambda x: x.split(' AND_REL ')[1],
        )
        relations_df['FeatureA'] = relations_df.FeatureA.map(
            lambda x: x.split(' AND_REL ')[0],
        )

        # redundancies include MI-scores of features w.r.t. non-label features
        redundancies_df = triplets[(
            triplets.FeatureB != args.label_column
        )].copy()
        redundancies_df = redundancies_df[
            redundancies_df.FeatureA !=
            args.label_column
        ]
        redundancies_df = redundancies_df[
            redundancies_df.apply(
                lambda x: (' AND_REL ' not in x.FeatureA)
                and (' AND_REL ' not in x.FeatureB),
                axis=1,
            )
        ]

        # normalize
        relevance_df['score'] = (relevance_df.Score - relevance_df.Score.min()) / (
            relevance_df.Score.max() - relevance_df.Score.min()
        )
        relations_df['score'] = (relations_df.Score - relations_df.Score.min()) / (
            relations_df.Score.max() - relations_df.Score.min()
        )
        redundancies_df['score'] = (
            redundancies_df.Score - redundancies_df.Score.min()
        ) / (redundancies_df.Score.max() - redundancies_df.Score.min())

        # create dicts
        relevance_dict = {
            row.FeatureA: row.score for _,
            row in relevance_df.iterrows()
        }
        relations_dict = {
            (row.FeatureA, row.FeatureB): row.score
            for _, row in relations_df.iterrows()
        }
        relations_dict.update(
            {
                (row.FeatureB, row.FeatureA): row.score
                for _, row in relations_df.iterrows()
            },
        )
        redundancy_dict = {
            (row.FeatureA, row.FeatureB): row.score
            for _, row in redundancies_df.iterrows()
        }

        # compute 3mr ranks
        mrmrmr_ranking = rank_features_3MR(
            relevance_dict, redundancy_dict, relations_dict,
        )
        mrmrmr_ranking.to_csv(
            os.path.join(args.output_folder, '3mr_ranks.tsv'), sep='\t', index=False,
        )

    feature_first_modified = []
    feature_second_modified = []

    if args.include_cardinality_in_feature_names == 'True':
        for enx in range(triplets.shape[0]):
            feature_first = triplets.iloc[enx]['FeatureA']
            feature_second = triplets.iloc[enx]['FeatureB']
            card_first = str(len(cardinality_object[feature_first]))
            card_second = str(len(cardinality_object[feature_second]))
            cov_first = int(
                round((np.mean(np.array(coverage_object[feature_first]))), 1),
            )
            cov_second = int(
                round(np.mean(np.array(coverage_object[feature_second])), 1),
            )

            feature_first_modified.append(
                feature_first + f'-({card_first}; {cov_first})',
            )
            feature_second_modified.append(
                feature_second + f'-({card_second}; {cov_second})',
            )

        triplets['FeatureA'] = feature_first_modified
        triplets['FeatureB'] = feature_second_modified

    feature_memory_df = pd.DataFrame(global_memory_storage).mean()
    feature_memory_df.columns = ['NormalizedSize']
    feature_memory_df.to_csv(
        f'{args.output_folder}/memory.tsv', sep='\t', index=True,
    )

    triplets = triplets.sort_values(by=['Score'])

    triplets.to_csv(
        os.path.join(args.output_folder, 'pairwise_ranks.tsv'), sep='\t', index=False,
    )

    # Write timings and config for replicability
    dfx = pd.DataFrame(all_timings)
    dfx.to_json(f'{args.output_folder}/timings.json')
    write_json_dump_to_file(args, f'{args.output_folder}/arguments.json')

    logging.info(
        f'Finished with ranking! Result stored as: {args.output_folder}/pairwise_ranks.tsv. Cleaning up tmp files ..',
    )

    os.remove('ranking_checkpoint_tmp.tsv')



from __future__ import annotations

from outrank.feature_transformations.feature_transformer_vault.default_transformers import DEFAULT_TRANSFORMERS
from outrank.feature_transformations.feature_transformer_vault.default_transformers import MINIMAL_TRANSFORMERS
from outrank.feature_transformations.feature_transformer_vault.fw_transformers import (
    FW_TRANSFORMERS,
)

_tr_global_namespace = {
    'default': DEFAULT_TRANSFORMERS,
    'minimal': MINIMAL_TRANSFORMERS,
    'fw-transformers': FW_TRANSFORMERS,
}



"""
This module implements probabilistic data structure which is able to calculate the cardinality of large multisets in a single pass using little auxiliary memory
"""
from __future__ import annotations

import numpy as np
import xxhash


class HyperLogLogWCache:
    def __init__(self, error_rate=0.005):
        # int(np.ceil(np.log2((1.04 / error_rate) ** 2)))
        self.p = 19
        self.m = 1 << self.p
        self.warmup_set = set()
        self.warmup_size = int(self.m / 2)
        self.width = 64 - self.p
        self.hll_flag = False

    def _hasher_update(self, value):
        self.hasher = xxhash.xxh32(seed=self.p)
        if isinstance(value, str):
            value = value.encode('utf-8')
            self.hasher.update(bytes(value))
        else:
            self.hasher.update(bytes(value))

        x = self.hasher.intdigest()
        j = x & (self.m - 1)
        w = x >> self.p

        rho = self.width - w.bit_length()
        self.M[j] = max(self.M[j], rho)

    def add(self, value):
        if len(self.warmup_set) < self.warmup_size and not self.hll_flag:
            self.warmup_set.add(value)
        elif not self.hll_flag:
            if not self.hll_flag:
                self.M = np.zeros(self.m)
                for element in self.warmup_set:
                    self._hasher_update(element)
                self.warmup_set = {}
            self.hll_flag = True
        else:
            self._hasher_update(value)

    def __len__(self):
        if self.hll_flag:
            basis = np.ceil(
                self.m *
                np.log(np.divide(self.m, len(np.where(self.M == 0)[0]))),
            )
            if basis != np.inf:
                return int(basis) - 1
            else:
                return 2**self.p
        else:
            return len(self.warmup_set)


if __name__ == '__main__':
    import random
    import string
    import time

    import matplotlib.pyplot as plt
    import pandas as pd
    import seaborn as sns
    import tqdm
    from pympler import asizeof

    def get_random_string(length):
        # choose from all lowercase letter
        letters = string.ascii_lowercase
        result_str = ''.join(random.choice(letters) for i in range(length))
        return result_str

    # results_df = []
    # num_vals = 100000
    # nbits = 16
    # for _ in range(3):
    #     for j in tqdm.tqdm(range(1000000, 10000000, 1000)):
    #         ground = list(set(np.random.randint(0, j, num_vals).tolist()))
    #         ground = ground + [
    #             get_random_string(random.randint(1, 15)) for k in range(j)
    #         ]

    #         start_time = time.time()
    #         GLOBAL_CARDINALITY_STORAGE = {}
    #         GLOBAL_CARDINALITY_STORAGE[1] = HyperLogLogWCache(0.005)

    #         for j in ground:
    #             GLOBAL_CARDINALITY_STORAGE[1].add(j)

    #         size1 = asizeof.asizeof(GLOBAL_CARDINALITY_STORAGE)
    #         error1 = 100 * \
    #             (1 - len(GLOBAL_CARDINALITY_STORAGE[1]) / len(set(ground)))
    #         end_time = time.time()
    #         tp1 = end_time - start_time

    #         import hyperloglog

    #         start_time = time.time()
    #         GLOBAL_CARDINALITY_STORAGE = {}
    #         GLOBAL_CARDINALITY_STORAGE[1] = hyperloglog.HyperLogLog(0.005)

    #         for j in ground:
    #             GLOBAL_CARDINALITY_STORAGE[1].add(j)
    #         size2 = asizeof.asizeof(GLOBAL_CARDINALITY_STORAGE)
    #         error2 = 100 * \
    #             (1 - len(GLOBAL_CARDINALITY_STORAGE[1]) / len(set(ground)))
    #         end_time = time.time()
    #         tp2 = end_time - start_time

    #         start_time = time.time()
    #         GLOBAL_CARDINALITY_STORAGE = set()

    #         for j in ground:
    #             GLOBAL_CARDINALITY_STORAGE.add(j)

    #         size3 = asizeof.asizeof(GLOBAL_CARDINALITY_STORAGE)
    #         error3 = 100 * \
    #             (1 - len(GLOBAL_CARDINALITY_STORAGE) / len(set(ground)))
    #         end_time = time.time()
    #         tp3 = end_time - start_time

    #         results_df.append(
    #             {
    #                 'num_samples': len(ground),
    #                 'time': tp3,
    #                 'algo': 'set',
    #                 'error': error3,
    #             },
    #         )
    #         results_df.append(
    #             {
    #                 'num_samples': len(ground),
    #                 'time': tp2,
    #                 'algo': 'default',
    #                 'error': error2,
    #             },
    #         )
    #         results_df.append(
    #             {
    #                 'num_samples': len(ground),
    #                 'time': tp1,
    #                 'algo': f'hllc ({nbits}, mixed)',
    #                 'error': error1,
    #             },
    #         )

    # out_df = pd.DataFrame(results_df)
    # out_df.to_csv('backup.csv')
    # print(out_df)
    # print(out_df.groupby('algo').mean())
    # sns.lineplot(
    #     x=out_df.num_samples, y=out_df.error,
    #     hue=out_df.algo, alpha=0.5,
    # )
    # plt.tight_layout()
    # plt.ylabel('Num. of unique values in data')
    # plt.ylabel('Abs error')
    # plt.savefig('linep.pdf')
    # plt.clf()
    # plt.cla()

    # sns.lineplot(
    #     x=out_df.num_samples.astype(
    #         float,
    #     ), y=out_df.time, hue=out_df.algo,
    # )
    # plt.tight_layout()
    # plt.ylabel('Time (s)')
    # plt.savefig('barp.pdf')
    # plt.clf()
    # plt.cla()



from __future__ import annotations

import logging
import os
from collections import defaultdict

import numpy as np
import pandas as pd

logging.basicConfig(format='%(asctime)s %(message)s', level=logging.INFO)


def outrank_task_result_summary(args):
    triplets = pd.read_csv(
        os.path.join(args.output_folder, 'pairwise_ranks.tsv'), sep='\t',
    )
    triplets = triplets.sort_values(by=['Score'], ascending=False)
    final_ranking = []
    for enx, row in triplets.iterrows():
        final_row = None
        if args.label_column == row['FeatureA'].split('-')[0]:
            final_row = [row['FeatureB'], row['Score']]
        if args.label_column == row['FeatureB'].split('-')[0]:
            final_row = [row['FeatureA'], row['Score']]
        if final_row and args.label_column != final_row[0]:
            final_ranking.append(final_row)

    final_df = pd.DataFrame(final_ranking)
    final_df.columns = ['Feature', f'Score {args.heuristic}']
    final_df.index = np.arange(1, final_df.shape[0] + 1, 1)
    final_df = (
        final_df.groupby(by=['Feature'])
        .median()
        .reset_index()
        .sort_values(by=[f'Score {args.heuristic}'], ascending=False)
    )

    min_score = np.min(final_df[f'Score {args.heuristic}'].values)
    max_score = np.max(final_df[f'Score {args.heuristic}'].values)
    final_df[f'Score {args.heuristic}'] = (
        final_df[f'Score {args.heuristic}'] - min_score
    ) / (max_score - min_score)
    logging.info(f'Storing summary files to {args.output_folder}')
    pd.set_option('display.max_rows', None, 'display.max_columns', None)
    singles_path = os.path.join(args.output_folder, 'feature_singles.tsv')
    final_df = final_df.reset_index(drop=True)
    final_df.to_csv(singles_path, sep='\t')

    if args.interaction_order > 1:
        feature_store = defaultdict(list)
        for enx, row in final_df.iterrows():
            fname = row['Feature']
            score = row[f'Score {args.heuristic}']
            if 'AND' in fname:
                for el in fname.split('-')[0].split(' AND '):
                    feature_store[el].append(score)

        final_aggregate_df = []
        for k, v in feature_store.items():
            final_aggregate_df.append(
                {
                    'Feature': k,
                    f'Combined score (order: {args.interaction_order}, {args.heuristic})': np.median(
                        v,
                    ),
                },
            )
        final_aggregate_df = pd.DataFrame(final_aggregate_df)
        final_aggregate_df.to_csv(
            os.path.join(args.output_folder, 'feature_singles_aggregated.tsv'), sep='\t',
        )

    final_df = final_df[final_df['Feature'].str.contains('_tr_')]
    final_df.to_csv(
        singles_path.replace('.tsv', '_transformers_only_imp.tsv'), sep='\t',
    )



from __future__ import annotations

import logging
import os

import pandas as pd

from outrank.visualizations.ranking_visualization import visualize_all

logging.basicConfig(format='%(asctime)s %(message)s', level=logging.INFO)


def outrank_task_visualize_results(args):
    logging.info(f'Beginning visualization based on: {args.output_folder}.')

    triplets = pd.read_csv(
        os.path.join(args.output_folder, 'pairwise_ranks.tsv'), sep='\t',
    )
    visualize_all(
        triplets,
        args.output_folder,
        args.label_column,
        args.reference_model_JSON,
        image_format=args.image_format,
        heuristic=args.heuristic,
    )



from __future__ import annotations

import numpy as np
from numba import njit
from numba import prange

np.random.seed(123)
# Fast Numba-based approximative mutual information


@njit(
    'Tuple((int32[:], int32[:]))(int32[:])',
    cache=True,
    fastmath=True,
    error_model='numpy',
    boundscheck=True,
)
def numba_unique(a):
    """Identify unique elements in an array, fast"""

    len_a = a.shape[0]
    container = np.zeros(np.max(a) + 1, dtype=np.int32)
    for el in range(len_a):
        container[a[el]] += 1

    unique_values = np.where(container != 0)[0]
    unique_counts = container[unique_values]
    return unique_values.astype(np.int32), unique_counts.astype(np.int32)


@njit(
    'float32(int32[:], int32[:], int32, float32)',
    cache=True,
    fastmath=True,
    error_model='numpy',
    boundscheck=True,
)
def compute_conditional_entropy(Y_classes, class_values, class_var_shape, initial_prob):
    conditional_entropy = 0.0

    for c in class_values:
        conditional_prob = np.count_nonzero(Y_classes == c) / class_var_shape
        if conditional_prob != 0:
            conditional_entropy -= (
                initial_prob * conditional_prob * np.log(conditional_prob)
            )

    return conditional_entropy


@njit(
    'float32(int32[:], int32[:], int32, int32[:], int32[:], b1)',
    cache=True,
    parallel=False,
    fastmath=True,
    error_model='numpy',
    boundscheck=True,
)
def compute_entropies(
    X, Y, all_events, f_values, f_value_counts, cardinality_correction,
):
    """Core entropy computation function"""

    conditional_entropy = 0.0
    background_cond_entropy = 0.0
    full_entropy = 0.0

    class_values, class_counts = numba_unique(Y)

    if not cardinality_correction:
        for k in prange(len(class_counts)):
            class_probability = class_counts[k] / all_events
            full_entropy += -class_probability * np.log(class_probability)

    for f_index in prange(len(f_values)):
        _f_value_counts = f_value_counts[f_index]

        if _f_value_counts == 1:
            continue

        initial_prob = _f_value_counts / all_events
        x_value_subspace = np.where(X == f_values[f_index])
        Y_classes = Y[x_value_subspace]
        conditional_entropy += compute_conditional_entropy(
            Y_classes, class_values, _f_value_counts, initial_prob,
        )

        if cardinality_correction:
            # A neat hack that seems to work fine (permutations are expensive)
            Y_classes = np.roll(Y, _f_value_counts)[x_value_subspace]

            background_cond_entropy += compute_conditional_entropy(
                Y_classes, class_values, _f_value_counts, initial_prob,
            )

    if not cardinality_correction:
        return full_entropy - conditional_entropy

    else:
        # note: full entropy falls out during derivation of final term
        core_joint_entropy = -conditional_entropy + background_cond_entropy
        return core_joint_entropy


@njit(
    'float32(int32[:], int32[:], float32, b1)',
    cache=True,
    fastmath=True,
    error_model='numpy',
    boundscheck=True,
)
def mutual_info_estimator_numba(
    Y, X, approximation_factor=1, cardinality_correction=False,
):
    """Core estimator logic. Compute unique elements, subset if required"""

    all_events = len(X)
    f_values, f_value_counts = numba_unique(X)

    # Diagonal entries
    if np.sum(X - Y) == 0:
        cardinality_correction = False

    if approximation_factor < 1:
        subspace_size = int(approximation_factor * all_events)
        if subspace_size != 0:
            subspace = np.random.randint(0, all_events, size=subspace_size)
            X = X[subspace]
            Y = Y[subspace]

    joint_entropy_core = compute_entropies(
        X, Y, all_events, f_values, f_value_counts, cardinality_correction,
    )

    return approximation_factor * joint_entropy_core


if __name__ == '__main__':
    import pandas as pd
    from sklearn.feature_selection import mutual_info_classif

    np.random.seed(123)
    import time

    final_times = []
    for algo in ['MI-numba-randomized']:
        for order in range(20, 21):
            for j in range(1):
                start = time.time()
                a = np.random.randint(1000, size=2**order).astype(np.int32)
                b = np.random.randint(1000, size=2**order).astype(np.int32)
                if algo == 'MI':
                    final_score = mutual_info_classif(
                        a.reshape(-1, 1), b.reshape(-1), discrete_features=True,
                    )
                elif algo == 'MI-numba-randomized':
                    final_score = mutual_info_estimator_numba(
                        a, b, np.float32(1.0), True,
                    )
                elif algo == 'MI-numba':
                    final_score = mutual_info_estimator_numba(
                        a, b, np.float32(1.0), False,
                    )
                elif algo == 'MI-numba-randomized-ap':
                    final_score = mutual_info_estimator_numba(
                        a, b, np.float32(0.3), True,
                    )
                elif algo == 'MI-numba-ap':
                    final_score = mutual_info_estimator_numba(
                        a, b, np.float32(0.3), False,
                    )

                end = time.time()
                tdiff = end - start
                instance = {
                    'time': tdiff,
                    'samples 2e': order, 'algorithm': algo,
                }
                final_times.append(instance)
                print(instance)
    dfx = pd.DataFrame(final_times)
    dfx = dfx.sort_values(by=['samples 2e'])
    print(dfx)



from __future__ import annotations

import logging
import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from scipy.cluster import hierarchy
from sklearn.manifold import TSNE
from sklearn.metrics import silhouette_score

from outrank.core_utils import read_reference_json

logging.basicConfig(format='%(asctime)s %(message)s', level=logging.INFO)
plt.rcParams['figure.figsize'] = (50, 30)


def visualize_hierarchical_clusters(
    triplet_dataframe: pd.DataFrame,
    output_folder: str,
    image_format: str = 'png',
    max_num_clusters: int = 100,
) -> None:
    """A method for visualization of hierarchical clusters w.r.t. different linkage functions"""

    # Prepare the canvas
    plt.rcParams['figure.figsize'] = (10, 5)
    unique_features = triplet_dataframe.FeatureA.unique()

    if len(unique_features) > 1000:
        logging.info('Trying to visualize too many features, exiting ..')
        exit()

    dmat = np.zeros((len(unique_features), len(unique_features)))
    logging.info('Preparing the data for clustering ..')

    if triplet_dataframe.shape[0] > 10**5:
        logging.info(
            'Trying to visualize more than 10 ** 5 triplets, exiting ..',
        )
        exit()

    pivot_table = pd.pivot_table(
        triplet_dataframe,
        values='Score',
        index='FeatureA',
        columns='FeatureB',
        aggfunc=np.mean,
    )

    # We need distances
    pivot_table.fillna(0, inplace=True)
    dmat = 1 - pivot_table.values

    # Visualize different dendrograms
    logging.info('Clustering ..')

    for linkage_heuristic in [
        # 'single', 'complete', 'average', 'weighted', 'centroid'
        'complete',
    ]:
        # Compute the linkage structure
        Z = hierarchy.linkage(dmat, linkage_heuristic)

        # Visualize
        hierarchy.dendrogram(
            Z, above_threshold_color='y', orientation='top', labels=unique_features,
        )
        # Store
        plt.title(f'Linkage function: {linkage_heuristic}')
        plt.tight_layout()
        out_path = f'{output_folder}/dendrogram_{linkage_heuristic}.{image_format}'
        plt.savefig(out_path, dpi=300)

        # Clean for subsequent plots
        plt.clf()
        plt.cla()
        logging.info(
            f'Visualized hierarchical clustering with linkage {linkage_heuristic} to {out_path}',
        )

        # Step 1: Identify relevant distance threshold bounds
        range_min, range_max = np.min(
            pivot_table.values,
        ), np.max(pivot_table.values)
        spectrum = np.arange(
            range_min, range_max,
            (range_max - range_min) / 1000,
        )
        max_silhouette = 0
        top_clustering = []
        full_silhouette_space = []

        # Step 2: Compute Silhouette for each threshold and store the results
        for possible_threshold in spectrum:
            cluster_assignments = hierarchy.fcluster(Z, possible_threshold)
            num_clusters = len(np.unique(cluster_assignments))
            if num_clusters > 2 and num_clusters < max_num_clusters:
                try:
                    sil_score = silhouette_score(
                        pivot_table, cluster_assignments,
                    )

                except Exception:
                    continue

                full_silhouette_space.append(
                    [
                        sil_score, possible_threshold, len(
                            np.unique(cluster_assignments),
                        ),
                    ],
                )
                if sil_score >= max_silhouette:
                    top_clustering = cluster_assignments
                    max_silhouette = sil_score

        # Step 3: We are interested in the best clustering w.r.t. Silhouette
        dfx = pd.DataFrame(full_silhouette_space)
        if len(dfx) == 0:
            logging.info('Silhouette space empty, exiting')
            exit()

        dfx.columns = ['Silhouette', 'threshold', 'numClusters']
        sns.lineplot(x=dfx.numClusters, y=dfx.Silhouette, color='black')
        plt.tight_layout()
        out_path = f'{output_folder}/SilhouetteProfile.{image_format}'
        plt.savefig(out_path, dpi=300)
        plt.clf()
        plt.cla()
        logging.info('Stored the Silhouette profile.')

        final_feature_cluster_df = pd.DataFrame(
            list(zip(top_clustering, pivot_table.index)),
        )
        final_feature_cluster_df.columns = ['ClusterID', 'Feature']
        final_feature_cluster_df.to_csv(
            f'{output_folder}/TopClustering.tsv', sep='\t',
        )

        # Get 2D embeddings of features and visualize them
        try:
            projected_data = TSNE().fit_transform(pivot_table.values)
            projected_data = pd.DataFrame(projected_data)
            projected_data.columns = ['Dim1', 'Dim2']
            projected_data['ClusterID'] = top_clustering
            projected_data['ClusterID'] = projected_data['ClusterID'].astype(
                str,
            )
            sns.scatterplot(
                x=projected_data.Dim1,
                y=projected_data.Dim2,
                hue=projected_data.ClusterID,
                palette='Set2',
            )
            plt.savefig(
                f'{output_folder}/clustersEmbeddingVisualization.pdf', dpi=300,
            )
            plt.clf()
            plt.cla()
        except:
            pass

        # Step 4: We are interested in the best clustering w.r.t. Silhouette
        # Not here yet

    plt.rcParams['figure.figsize'] = (50, 30)


def visualize_heatmap(
    triplets: pd.DataFrame, output_folder: str, image_format: str,
) -> None:
    # Compute the interaction pivot table
    sns.set(font_scale=2)
    fig, ax = plt.subplots()
    pivot_table = pd.pivot_table(
        triplets, values='Score', index='FeatureA', columns='FeatureB', aggfunc=np.mean,
    )
    mask = np.zeros_like(pivot_table.values)
    mask[np.triu_indices_from(mask)] = True
    fsize_heatmap = 20
    if pivot_table.shape[0] > 100:
        sns.set(font_scale=1)
        fsize_heatmap = 3

    logging.info('Visualizing the heatmap ..')

    if pivot_table.shape[0] > 500:
        logging.info(
            'Skipping heatmap visualization due to too many elements ..',
        )
        return

    # Visualize the table
    plt.figure(figsize=(50, 50))
    plt.rcParams.update({'font.size': 1})
    sns.heatmap(
        pivot_table,
        annot=True,
        mask=mask,
        annot_kws={'size': fsize_heatmap},
        square=False,
        cmap='coolwarm',
        linecolor='black',
        linewidths=0.05,
    )
    plt.xlabel('')
    plt.ylabel('')
    plt.tight_layout()
    plt.savefig(f'{output_folder}/heatmap.{image_format}', dpi=500)
    plt.clf()
    plt.cla()
    logging.info(f'Stored heatmap to: {output_folder}/heatmap.{image_format}')


def visualize_barplots(
    triplets: pd.DataFrame,
    output_folder: str,
    reference_json: str,
    image_format: str,
    label: str,
    heuristic: str,
) -> None:
    # Extract only the interactions related to the target attribute
    sns.set(font_scale=8)
    feature_ranks_rows = []
    for enx, row in triplets.iterrows():
        feature_A = row['FeatureA']
        feature_B = row['FeatureB']
        if label in feature_A:
            feature_ranks_rows.append([feature_B, row.Score])
        elif label in feature_B:
            feature_ranks_rows.append([feature_A, row.Score])

    # Align with an existing model
    feature_ranks: pd.DataFrame = pd.DataFrame(feature_ranks_rows)
    feature_ranks.columns = ['Feature', 'Value']
    feature_ranks = feature_ranks[
        ~feature_ranks['Feature'].str.contains(
            label,
        )
    ]
    if not os.path.exists(reference_json):
        reference_json = ''

    if reference_json:
        ref_json = read_reference_json(reference_json)
        used_features = []
        if 'features' in ref_json['desc']:
            for feature in ref_json['desc']['features']:
                used_features.append(feature)

        if 'fields' in ref_json['desc']:
            for field in ref_json['desc']['fields']:
                used_features.append(field)
    else:
        used_features = feature_ranks.keys()

    feature_ranks['Feature'] = feature_ranks['Feature'].astype(str)
    feature_ranks['Value'] = feature_ranks['Value'].astype(float)
    feature_ranks = feature_ranks.groupby(
        by=['Feature'],
    ).median().reset_index()
    feature_ranks = feature_ranks.sort_values(by=['Value'], ascending=False)

    subset_ranges = [10, 25, 50, 100, feature_ranks.shape[0]]
    sns.set_style('whitegrid')

    for subset_range in subset_ranges:
        feature_ranks_reduced = feature_ranks.copy().iloc[:subset_range]
        plt.figure(figsize=(18, 12))
        fig, ax = plt.subplots()

        if (
            feature_ranks_reduced.shape[0] > 45
            and feature_ranks_reduced.shape[0] <= 100
        ):
            ax.yaxis.set_tick_params(labelsize=8)
        elif feature_ranks_reduced.shape[0] > 100:
            ax.yaxis.set_tick_params(labelsize=2)
        else:
            ax.yaxis.set_tick_params(labelsize=25)

        # Visualize the barplot
        plt.title(f'Ranking w.r.t "{label}"\n')
        sns.barplot(
            x='Value',
            y='Feature',
            errwidth=0.7,
            data=feature_ranks_reduced,
            palette='coolwarm_r',
        )

        # Modify the ticks if needed
        for item in ax.get_yticklabels():
            for prod_feature in used_features:
                if item.get_text() in prod_feature:
                    item.set_fontweight('bold')
                    item.set_color('red')
                    break

        plt.xlabel(f'Feature importance (based on heuristic {heuristic})')
        plt.ylabel('')
        plt.tight_layout()
        plt.savefig(
            f'{output_folder}/barplot_top_{subset_range}.{image_format}', dpi=300,
        )
        plt.clf()
        plt.cla()

        logging.info(
            f'Stored barplot to: {output_folder}/barplot_top_{subset_range}_.{image_format}',
        )


def visualize_all(
    triplets: pd.DataFrame,
    output_folder: str,
    label: str = '',
    reference_json: str = '',
    image_format: str = 'png',
    heuristic: str = 'MI',
) -> None:
    """A method for visualization of the obtained feature interaction maps."""

    if not os.path.exists(output_folder):
        os.makedirs(output_folder)

    # Visualize feature clusters
    visualize_hierarchical_clusters(triplets, output_folder, image_format)

    # Visualize heatmap
    visualize_heatmap(triplets, output_folder, image_format)

    # visualize barplot
    visualize_barplots(
        triplets, output_folder, reference_json, image_format, label, heuristic,
    )



# A collection of feature transformers that can be considered
from __future__ import annotations

import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Set

import numpy as np
import pandas as pd

import outrank.feature_transformations.feature_transformer_vault as transformer_vault
from outrank.core_utils import internal_hash

logging.basicConfig(format='%(asctime)s %(message)s', level=logging.INFO)


class FeatureTransformerNoise:
    def __init__(self):
        self.noise_preset = 'default'

    def construct_new_features(self, dataframe: pd.DataFrame, label_column=None):
        """Generate a few standard noise distributions"""

        new_columns = dict()
        if self.noise_preset == 'default':
            new_columns['CONTROL-constant0'] = np.array([0] * dataframe.shape[0])
            new_columns['CONTROL-gaussian'] = np.random.normal(
                size=dataframe.shape[0],
            )
            new_columns['CONTROL-uniform'] = np.random.random(
                dataframe.shape[0],
            )
            new_columns['CONTROL-random-binary'] = np.random.randint(
                0, 2, dataframe.shape[0],
            )
            new_columns['CONTROL-random-card100'] = np.random.randint(
                0, 1 + 1 * 10**2, dataframe.shape[0],
            )
            new_columns['CONTROL-random-card2k'] = np.random.randint(
                0, 1 + 2 * 10**3, dataframe.shape[0],
            )
            new_columns['CONTROL-random-card10k'] = np.random.randint(
                0, 1 + 10 * 10**3, dataframe.shape[0],
            )
            new_columns['CONTROL-random-card50k'] = np.random.randint(
                0, 1 + 50 * 10**3, dataframe.shape[0],
            )
            new_columns['CONTROL-int-sequence'] = np.arange(
                0, dataframe.shape[0], 1.0,
            )

            if label_column not in dataframe.columns:
                logging.warn(
                    'Could not find target feature in your data set - please inspect the columns if doing targeted ranking!',
                )
            else:
                new_columns['CONTROL-target'] = dataframe[label_column]

            new_columns['CONTROL-volume'] = np.array([
                internal_hash(str(x)) for _, x in dataframe.iterrows()
            ])
        else:
            # Not relevant yet; will be if this is useful.
            pass

        if len(new_columns) > 0:
            tmp_df = pd.DataFrame(new_columns)
            dataframe = pd.concat([dataframe, tmp_df], axis=1)
            del tmp_df

        return dataframe


class FeatureTransformerGeneric:
    def __init__(self, numeric_column_names: set[str], preset: str = 'default'):
        for transformer_namespace in preset.split(','):
            self.transformer_collection: dict[str, str] = dict()
            transformer_subspace = transformer_vault._tr_global_namespace.get(
                transformer_namespace, None,
            )
            if transformer_subspace:
                self.transformer_collection = {
                    **self.transformer_collection,
                    **transformer_subspace,
                }

            if len(self.transformer_collection) == 0:
                raise NotImplementedError(
                    'Please, specify valid transformer namespaces (e.g., default, minimal etc.)',
                )

        self.numeric_column_names = set(numeric_column_names)
        self.constructed_feature_names: set[str] = set()

        # If 80% of values are the same, don't consider a transformation
        self.max_maj_support = 0.80

        # If more than 75% of vals are missing, don't consider a transformation
        self.nan_prop_support = 0.75

    def get_vals(self, tmp_df: pd.DataFrame, col_name: str) -> Any:
        cvals = tmp_df[col_name].values.tolist()
        cvals = [str(x).replace('"', '') for x in cvals]
        cvals = [0.0 if len(x) == 0 else float(x) for x in cvals]

        return np.array(cvals)

    def construct_baseline_features(self, dataframe: Any) -> pd.DataFrame:
        fvals = []
        for enx, row in dataframe.iterrows():
            missing_prop = np.round(
                row.values.tolist().count('') / dataframe.shape[1], 1,
            )
            fvals.append(missing_prop)

        dataframe['BASELINE-MISSING-PROPORTION'] = fvals
        dataframe['BASELINE-DUMMY'] = 0

        return dataframe

    def construct_new_features(self, dataframe: Any) -> pd.DataFrame:
        new_numeric = set()
        logging.info(
            f'Considering {len(self.transformer_collection)} transformations for {len(self.numeric_column_names)} features ({len(self.transformer_collection) * len(self.numeric_column_names)} new features will be considered).',
        )

        invalid_transforms = 0
        new_columns = dict()
        for numeric_column in self.numeric_column_names:
            X = self.get_vals(dataframe, numeric_column)

            if len(X) == 0:
                raise AssertionError(
                    f"Could not retrieve the colomn {numeric_column}'s values. Please check the data.",
                )

            for k, v in self.transformer_collection.items():
                feature_name = f'{numeric_column}{k}'
                transformed_array = eval(v).astype(str)
                u, c = np.unique(transformed_array, return_counts=True)
                nan_prop = np.count_nonzero(transformed_array == 'nan') / len(
                    transformed_array,
                )
                cfreq = np.divide(np.max(c), np.sum(c))
                if (
                    len(u) > 1
                    and cfreq < self.max_maj_support
                    and nan_prop < self.nan_prop_support
                ):
                    new_columns[feature_name] = transformed_array
                    new_numeric.add(feature_name)

                else:
                    invalid_transforms += 1

        if len(new_columns) > 0:
            tmp_df = pd.DataFrame(new_columns)
            dataframe = pd.concat([dataframe, tmp_df], axis=1)
            del tmp_df

        logging.info(
            f'{invalid_transforms} invalid transformations were skipped.',
        )
        self.numeric_column_names = self.numeric_column_names
        self.constructed_feature_names = new_numeric
        return dataframe



from __future__ import annotations

import argparse
import json
import logging

from outrank.task_generators import outrank_task_generate_data_set
from outrank.task_ranking import outrank_task_conduct_ranking
from outrank.task_selftest import conduct_self_test
from outrank.task_summary import outrank_task_result_summary
from outrank.task_visualization import outrank_task_visualize_results

logging.basicConfig(
    format='%(asctime)s - %(message)s',
    datefmt='%d-%b-%y %H:%M:%S',
)
logging.getLogger(__name__).setLevel(logging.INFO)

usage_examples = """
    Usage examples:

    # perform ranking, summary and visualize the results
    outrank --task all --data_path pathToSomeData --data_source ob-vw --heuristic MI-numba-randomized --include_cardinality_in_feature_names True --target_ranking_only True --combination_number_upper_bound 2048 --num_threads 8 --interaction_order 1 --transformers fw-transformers --output_folder ./ranking_outputs --subsampling 100

    # pairwise ranking only
    outrank --task ranking --data_path pathToSomeData --data_source ob-vw --heuristic MI-numba-randomized --target_ranking_only False --combination_number_upper_bound 10000 --num_threads 30 --output_folder ./ranking_outputs --subsampling 10

    # Higher order interactions
    outrank --task all --data_path pathToSomeData --data_source csv-raw --heuristic MI-numba-randomized --target_ranking_only True --combination_number_upper_bound 2048 --num_threads 8 --interaction_order 3 --output_folder ./ranking_outputs --subsampling 20

    # More docs and use cases at https://improved-dollop-9k1wgvm.pages.github.io/outrank.html
"""


def main():
    parser = argparse.ArgumentParser(
        description='Fast feature screening for sparse data sets.',
        epilog=usage_examples,
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        '--task',
        type=str,
        default='all',
        help='Type of task to consider. Can be either "ranking", "ranking_summary", "feature_summary_transformers",  or "visualization"',
    )

    parser.add_argument(
        '--minibatch_size',
        type=int,
        default=2**13,
        help='Suitable for data, not pre-split to batches, this parameter determines batch size - note that too large batch size can slow down the multithreaded score computation due to many thread allocations etc. This works ok for <300 features and up to 48 threads.',
    )

    parser.add_argument(
        '--output_folder',
        type=str,
        default='ranking_outputs',
        help='Output folder containing ranking results.',
    )

    parser.add_argument(
        '--data_source',
        type=str,
        default='ob-vw',
        help='Which database is used to obtain learning instances? this determines the inferred folder structure (csv-raw, ob-vw, ob-csv).',
    )

    parser.add_argument(
        '--data_path',
        type=str,
        default=None,
        help='Path to the folder containing the main data used for subsequent learning.',
    )

    parser.add_argument(
        '--subsampling',
        type=int,
        default=10,
        help='Subsampling ratio - every n-th instance will be considered (suggested value: 10 to 100)',
    )

    parser.add_argument(
        '--combination_number_upper_bound',
        type=int,
        default=2**15,
        help='Cap the number of generated combinations (random sampling of that space).',
    )

    parser.add_argument(
        '--missing_value_symbols',
        type=str,
        default=',{}',
        help='What symbols denote missing values? Comma-separate them.',
    )

    parser.add_argument(
        '--heuristic',
        type=str,
        default='MI-numba-randomized',
        help='Selected heuristic (that performs feature scoring). For full list please see the docs: https://improved-dollop-9k1wgvm.pages.github.io/outrank/algorithms/importance_estimator.html#get_importances_estimate_pairwise',
    )

    parser.add_argument(
        '--include_noise_baseline_features',
        type=str,
        default='False',
        help='If enabled, it computes five control variables (random noises)',
    )

    parser.add_argument(
        '--include_cardinality_in_feature_names',
        type=str,
        default='True',
        help='If enabled, feature names appear as feature-(cardinality) for easier inspection/debugging.',
    )

    parser.add_argument(
        '--image_format',
        type=str,
        default='pdf',
        help='The format of the output images (task: visualization)',
    )

    parser.add_argument(
        '--num_threads', type=int, default=8, help='Number of threads to consider',
    )

    parser.add_argument(
        '--label_column',
        type=str,
        default='label',
        help='Name of the target attribute (useful for visualization)',
    )

    parser.add_argument(
        '--transformers',
        type=str,
        default='none',
        help='Collection of which feature transformations to consider',
    )

    parser.add_argument(
        '--rare_value_count_upper_bound',
        type=int,
        default=1,
        help="When identifying rare attr-val pairs, what's the upper frequency bound?",
    )

    parser.add_argument(
        '--feature_set_focus',
        type=str,
        default=None,
        help='Collection of which feature transformations to consider',
    )

    parser.add_argument(
        '--interaction_order',
        type=int,
        default=1,
        help='The order of feature interactions to consider during ranking (complex features comprised of n elementary ones)',
    )

    parser.add_argument(
        '--reference_model_JSON',
        type=str,
        default='./ranking_outputs/reference_model.json',
        help='Reference model JSON',
    )

    parser.add_argument(
        '--target_ranking_only',
        type=str,
        default='True',
        help='Compute only the feature-label scores? This is substantially faster (O(n)).',
    )

    parser.add_argument(
        '--explode_multivalue_features',
        type=str,
        default='False',
        help="Which ';'-separated features should be one-hot encoded into n new features (coverage analysis)",
    )

    parser.add_argument(
        '--subfeature_mapping',
        type=str,
        default='False',
        help='Compute sub-features on-the fly. Example: featureA->featureB implies features based on each value of featureA will be considered.',
    )

    parser.add_argument(
        '--num_synthetic_features',
        type=int,
        default=100,
        help='Relevant for task data_generator -- how many features.',
    )

    parser.add_argument(
        '--num_synthetic_rows',
        type=int,
        default=1000000,
        help='Relevant for task data_generator -- how many rows.',
    )

    parser.add_argument(
        '--generator_type',
        type=str,
        default='naive',
        help='Relevant for task data_generator -- which generator to consider',
    )

    parser.add_argument(
        '--output_synthetic_df_name',
        type=str,
        default='test_data_synthetic',
        help='Relevant for task data_generator -- name of the folder that contains generated data.',
    )

    args = parser.parse_args()

    if args.task == 'selftest':
        conduct_self_test()
        exit()

    if args.data_path is None and args.task != 'data_generator':
        logging.error('Please specify data set name (--data_path).')
        exit()

    all_tasks_to_consider = []
    if args.task != 'all':
        all_tasks_to_consider = [args.task]

    else:
        all_tasks_to_consider = ['ranking', 'ranking_summary', 'visualization']

    for task in all_tasks_to_consider:
        logging.info(f'Proceeding with task: {task} ..')

        if (
            task == 'ranking'
            or task == 'feature_summary_transformers'
            or task == 'identify_rare_values'
        ):
            outrank_task_conduct_ranking(args)

        elif task == 'visualization':
            outrank_task_visualize_results(args)

        elif task == 'ranking_summary':
            outrank_task_result_summary(args)

        elif task == 'data_generator':
            outrank_task_generate_data_set(args)

        else:
            logging.info(f'Warning, the selected task: {task} does not exist.')


if __name__ == '__main__':
    main()



from __future__ import annotations

import numpy as np

from outrank.feature_transformations.feature_transformer_vault.default_transformers import (
    DEFAULT_TRANSFORMERS,
)

FW_TRANSFORMERS = DEFAULT_TRANSFORMERS.copy()
resolution_range = [1, 10, 50, 100]
greater_than_range = [1, 2, 4, 8, 16, 32, 64, 96]

for resolution in resolution_range:
    for greater_than in greater_than_range:
        FW_TRANSFORMERS[f'_tr_fw_sqrt_res_{resolution}_gt_{greater_than}'] = (
            f'np.where(X < {greater_than}, '
            f'X, '
            f'np.where(X>{greater_than} ,'
            f'np.round(np.sqrt(X-{greater_than})*{resolution},0), 0))'
        )

        FW_TRANSFORMERS[
            f'_tr_fw_log_res_{resolution}_gt_{greater_than}'
        ] = f'np.where(X <{greater_than}, X, np.where(X >{greater_than}, np.round(np.log(X-{greater_than})*{resolution},0), 0))'

for resolution in resolution_range:
    for greater_than in [np.divide(x, 100) for x in greater_than_range]:
        FW_TRANSFORMERS[
            f'_tr_fw_prob_sqrt_res_{resolution}_gt_{greater_than}'
        ] = f'np.where(X < {greater_than}, X, np.where(X>{greater_than}, np.round(np.sqrt(X-{greater_than})*{resolution},0), 0))'

        FW_TRANSFORMERS[
            f'_tr_fw_prob_log_res_{resolution}_gt_{greater_than}'
        ] = f'np.where(X <{greater_than},X, np.where(X>{greater_than}, np.round(np.log(X-{greater_than})*{resolution},0), 0))'

if __name__ == '__main__':
    print(len(FW_TRANSFORMERS))



# helper set of methods that enable anywhere verification of core functions
from __future__ import annotations

import logging
import os
import shutil
import subprocess

import pandas as pd

logging.basicConfig(
    format='%(asctime)s - %(message)s',
    datefmt='%d-%b-%y %H:%M:%S',
)
logger = logging.getLogger('syn-logger')
logger.setLevel(logging.DEBUG)


def conduct_self_test():
    # Simulate full flow, ranking only
    subprocess.run(
        'outrank --task data_generator --num_synthetic_rows 100000', shell=True,
    )
    subprocess.run(
        'outrank --task ranking --data_path test_data_synthetic --data_source csv-raw;',
        shell=True,
    )

    dfx = pd.read_csv('ranking_outputs/pairwise_ranks.tsv', sep='\t')

    logger.info("Verifying output's properties ..")
    assert dfx.shape[0] == 201
    assert dfx.shape[1] == 3
    assert dfx['FeatureA'].values.tolist().pop() == 'label-(81; 100)'

    to_remove = ['ranking_outputs', 'test_data_synthetic']
    for path in to_remove:
        if os.path.exists(path) and os.path.isdir(path):
            logger.info(f'Removing {path} as part of cleanup ..')
            shutil.rmtree(path)

    logger.info('All tests passed, OutRank seems in shape \N{winking face}')



# This simplest thing we can do for now.
from __future__ import annotations

import numpy as np

np.random.seed(123)


def generate_random_matrix(num_features=100, size=20000):
    # random int matrix (categorical)
    sample = np.random.randint(10, 100, size=(size, num_features))

    target = sample[:, 30]
    # Some noise

    target[target < 20] = 0
    return sample, target


if __name__ == '__main__':
    import argparse
    import logging
    import os
    import shutil

    import pandas as pd

    logging.basicConfig(
        format='%(asctime)s - %(message)s',
        datefmt='%d-%b-%y %H:%M:%S',
    )
    logger = logging.getLogger('syn-logger')
    logger.setLevel(logging.DEBUG)

    parser = argparse.ArgumentParser(
        description='Fast feature screening for sparse data sets.',
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument('--output_df_name', type=str, default=None)

    parser.add_argument('--verify_outputs', type=str, default=None)

    parser.add_argument('--num_features', type=int, default=300)

    parser.add_argument('--size', type=int, default=1000)

    args = parser.parse_args()

    if args.output_df_name is not None:
        sample, target = generate_random_matrix(args.num_features, args.size)
        dfx = pd.DataFrame(sample)
        dfx.columns = [f'f{x}' for x in range(dfx.shape[1])]
        dfx['label'] = target
        if os.path.exists(args.output_df_name) and os.path.isdir(args.output_df_name):
            shutil.rmtree(args.output_df_name)
        os.mkdir(args.output_df_name)
        dfx.to_csv(f'./{args.output_df_name}/data.csv', index=False)

        logging.info(f'Generated dataset {dfx.shape} in {args.output_df_name}')
    elif args.verify_outputs is not None:
        rankings = pd.read_csv(
            os.path.join(args.verify_outputs, 'feature_singles.tsv'), sep='\t',
        )
        if rankings.iloc[1]['Feature'] != 'f30-(81; 100)':
            raise Exception(
                f'Could not retrieve the appropriate feature needle in the haystack {rankings.iloc[1].Feature}, exiting',
            )
        else:
            logger.info(
                f'Identified the appropriate feature in the haystack ({rankings.iloc[1].Feature})',
            )



# Some boilerplate transformations people tend to use
from __future__ import annotations
MINIMAL_TRANSFORMERS = {
    '_tr_sqrt': 'np.sqrt(X)',
    '_tr_log(x+1)': 'np.log(X + 1)',
    '_tr_sqrt(abs(x))': 'np.sqrt(np.abs(X))',
    '_tr_log(abs(x)+1)': 'np.log(np.abs(X) + 1)',
}

DEFAULT_TRANSFORMERS = {
    '_tr_sqrt': 'np.sqrt(X)',
    '_tr_log(x+1)': 'np.log(X + 1)',
    '_tr_sqrt(abs(x))': 'np.sqrt(np.abs(X))',
    '_tr_log(abs(x)+1)': 'np.log(np.abs(X) + 1)',
    '_tr_div(x,abs(x))*log(abs(x))': 'np.divide(X, np.abs(X)) * np.log(np.abs(X))',
    '_tr_log(x + sqrt(pow(x,2), 1)': 'np.log(X + np.sqrt(np.power(X, 2) + 1))',
    '_tr_log*sqrt': 'np.log(X + 1) * np.sqrt(X)',
    '_tr_log*100': 'np.round(np.log(X + 1) * 100, 0)',
    '_tr_nonzero': 'np.where(X != 0, 1, 0)',
    '_tr_round(div(x,max))': 'np.round(np.divide(X, np.max(X)), 0)',
}

if __name__ == '__main__':
    import numpy as np

    # generate some input (call it X)
    X = np.random.random(100)

    # get some transformer
    some_transformer = DEFAULT_TRANSFORMERS.get('_tr_nonzero')

    if some_transformer is None:
        some_transformer = ''

    # evaluate to get output
    output = eval(some_transformer)

    # check output somehow
    print(output)



from __future__ import annotations

import os

import setuptools


def _parse_requirements(file):
    required_packages = []
    with open(
        os.path.join(os.path.dirname(__file__), file), encoding='utf-8',
    ) as req_file:
        for line in req_file:
            required_packages.append(line.strip())
    return required_packages


def _read_description():
    with open('README.md') as description:
        return description.read()


packages = [x for x in setuptools.find_packages() if x != 'test']
setuptools.setup(
    name='outrank',
    version='0.91',
    description='OutRank: Feature ranking for massive sparse data sets.',
    long_description=_read_description(),
    long_description_content_type='text/markdown',
    url='https://github.com/outbrain/outrank',
    author='Research Infra (Outbrain); Blaz Skrlj led the development of this project',
    license='BSD',
    entry_points={'console_scripts': ['outrank = outrank.__main__:main']},
    packages=packages,
    zip_safe=True,
    include_package_data=True,
    install_requires=_parse_requirements('requirements.txt'),
)



# A module for pairwise computation of importances -- entrypoint for the core ranking algorighm(s)
from __future__ import annotations

import operator
import traceback
from typing import Any
from typing import Dict

import numpy as np
import pandas as pd
from scipy.stats import pearsonr
from sklearn.feature_selection import mutual_info_classif
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import adjusted_mutual_info_score
from sklearn.model_selection import cross_val_score
from sklearn.preprocessing import OneHotEncoder
from sklearn.svm import SVC

try:
    from outrank.algorithms.feature_ranking import ranking_mi_numba

    numba_available = True

except Exception as es:
    traceback.print_exc(0)
    numba_available = False


def sklearn_MI(vector_first: Any, vector_second: Any) -> float:
    estimate_feature_importance = mutual_info_classif(
        vector_first.reshape(-1, 1), vector_second.reshape(-1), discrete_features=True,
    )[0]
    return estimate_feature_importance


def sklearn_surrogate(
    vector_first: Any, vector_second: Any, surrogate_model: str,
) -> float:
    if surrogate_model == 'surrogate-LR':
        clf = LogisticRegression(max_iter=100000)
    elif surrogate_model == 'surrogate-SVM':
        clf = SVC(gamma='auto', probability=True)

    transf = OneHotEncoder()

    # They do not commute, swap if needed
    if len(np.unique(vector_second) > 2):
        vector_third = vector_second
        vector_second = vector_first
        vector_first = vector_third
        del vector_third

    unique_values, counts = np.unique(vector_second, return_counts=True)

    # Establish min support for this type of ranking.
    if counts[0] < len(unique_values) * (2**5):
        estimate_feature_importance = 0

    else:
        vector_first = transf.fit_transform(vector_first.reshape(-1, 1))
        estimate_feature_importance_list = cross_val_score(
            clf, vector_first, vector_second, scoring='neg_log_loss', cv=4,
        )

        estimate_feature_importance = 1 + \
            np.median(estimate_feature_importance_list)

    return estimate_feature_importance


def numba_mi(vector_first, vector_second, heuristic):
    if heuristic == 'MI-numba-randomized':
        cardinality_correction = True

    else:
        cardinality_correction = False

    estimate_feature_importance = ranking_mi_numba.mutual_info_estimator_numba(
        vector_first.reshape(-1).astype(np.int32),
        vector_second.reshape(-1).astype(np.int32),
        approximation_factor=np.float32(1.0),
        cardinality_correction=cardinality_correction,
    )

    return estimate_feature_importance


def sklearn_mi_adj(vector_first, vector_second):
    # AMI(U, V) = [MI(U, V) - E(MI(U, V))] / [avg(H(U), H(V)) - E(MI(U, V))]
    estimate_feature_importance = adjusted_mutual_info_score(
        vector_first.reshape(-1), vector_second.reshape(-1),
    )
    return estimate_feature_importance


def get_importances_estimate_pairwise(combination, args, tmp_df):
    """A method for parallel importances estimation. As interaction scoring is independent, individual scores can be computed in parallel."""

    feature_one = combination[0]
    feature_two = combination[1]

    vector_first = tmp_df[[feature_one]].values.ravel()
    vector_second = tmp_df[[feature_two]].values.ravel()

    if len(vector_first) == 0 or len(vector_second) == 0:
        return [feature_one, feature_two, 0]

    # Compute score based on the selected heuristic.
    if args.heuristic == 'MI':
        # Compute the infoGain
        estimate_feature_importance = sklearn_MI(vector_first, vector_second)

    elif 'surrogate-' in args.heuristic:
        estimate_feature_importance = sklearn_surrogate(
            vector_first, vector_second, args.heuristic,
        )

    elif 'MI-numba' in args.heuristic:
        estimate_feature_importance = numba_mi(
            vector_first, vector_second, args.heuristic,
        )

    elif args.heuristic == 'AMI':
        estimate_feature_importance = sklearn_mi_adj(
            vector_first, vector_second,
        )

    elif args.heuristic == 'correlation-Pearson':
        estimate_feature_importance = pearsonr(vector_first, vector_second)[0]

    elif args.heuristic == 'Constant':
        estimate_feature_importance = 0.0

    else:
        raise ValueError(
            'Please select one of the possible heuristics (MI, chi2)',
        )

    return (feature_one, feature_two, estimate_feature_importance)


def rank_features_3MR(
    relevance_dict: dict[str, float],
    redundancy_dict: dict[tuple[Any, Any], Any],
    relational_dict: dict[tuple[Any, Any], Any],
    strategy: str = 'median',
    alpha: float = 1,
    beta: float = 1,
) -> pd.DataFrame:
    all_features = relevance_dict.keys()
    most_important_feature = max(
        relevance_dict.items(), key=operator.itemgetter(1),
    )[0]
    ranked_features = [most_important_feature]

    def calc_higher_order(feature, is_redundancy=True):
        values = []
        for feat in ranked_features:
            if is_redundancy:
                values.append(redundancy_dict[(feat, feature)])
            else:
                values.append(relational_dict[(feat, feature)])
        if strategy == 'sum':
            return sum(values)
        if strategy == 'mean':
            return np.mean(values)
        return np.median(values)

    while len(ranked_features) != len(all_features):
        top_importance = 0
        most_important_feature = ''

        for ind, feat in enumerate(set(all_features) - set(ranked_features)):
            feature_redundancy = calc_higher_order(feat)
            feature_relation = calc_higher_order(feat, False)
            feature_relevance = relevance_dict[feat]
            importance = (
                feature_relevance - alpha * feature_redundancy + beta * feature_relation
            )

            if (importance > top_importance) or (ind == 0):
                top_importance = importance
                most_important_feature = feat
        ranked_features.append(most_important_feature)
    return pd.DataFrame(
        {
            'Feature': ranked_features,
            '3mr_ranking': list(range(1, len(ranked_features) + 1)),
        },
    )


def get_importances_estimate_nonmyopic(args: Any, tmp_df: pd.DataFrame):
    # TODO - nonmyopic algorithms - tmp_df \ args.label vs. label
    # TODO - this is to be executed directly on df - no need for parallel kernel(s)
    pass



# OutRank is also capable of generating data sets.
from __future__ import annotations

import logging
import os
import shutil

import pandas as pd

from outrank.algorithms.synthetic_data_generators import generator_naive

logging.basicConfig(
    format='%(asctime)s - %(message)s',
    datefmt='%d-%b-%y %H:%M:%S',
)
logger = logging.getLogger('syn-logger')
logger.setLevel(logging.DEBUG)


def outrank_task_generate_data_set(args):
    """Core method for generating data sets"""

    if args.generator_type == 'naive':
        sample, target = generator_naive.generate_random_matrix(
            args.num_synthetic_features, args.num_synthetic_rows,
        )
    else:
        raise ValueError(f'Generator {args.generator_type} not implemented.')

    dfx = pd.DataFrame(sample)
    dfx.columns = [f'f{x}' for x in range(dfx.shape[1])]
    dfx['label'] = target
    if os.path.exists(args.output_synthetic_df_name) and os.path.isdir(
        args.output_synthetic_df_name,
    ):
        logger.info(
            f'Found existing: {args.output_synthetic_df_name}, removing first ..',
        )
        shutil.rmtree(args.output_synthetic_df_name)
    os.mkdir(args.output_synthetic_df_name)
    dfx.to_csv(f'./{args.output_synthetic_df_name}/data.csv', index=False)

    logger.info(
        f'Generated data set of shape {dfx.shape} in {args.output_synthetic_df_name}',
    )
